/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.1, (the "License"); you may not use this file except in
 * compliance with the License. You should have received a copy of the
 * Erlang Public License along with this software. If not, it can be
 * retrieved via the world wide web at http://www.erlang.org/.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Original Code is Scott Lystig Fritchie.
 * Portions created by Mr. Fritchie are Copyright 2003, Scott Lystig
 * Fritchie, <slfritchie@snookles.com>. All Rights Reserved.''
 * 
 *     $Id$
 */
/*
** Implementation of JudySL-based ETS tables.
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "sys.h"
#include "erl_vm.h"
#include "global.h"
#include "erl_process.h"
#include "error.h"
#include "erl_db.h"
#include "bif.h"
#include "big.h"
#include "dist.h"
#include "external.h"

/* XXX Fix these included to be done Correctly */
#include	<stdio.h>
#include	<string.h>
#include	<sys/param.h>
#include	<sys/time.h> /* XXX debugging only */

#ifdef DEBUG
#include <stdarg.h>
#define HARDDEBUG 1
#endif	/* DEBUG */

#include "erl_db_judy.h"
#include "erl_db_util.h"

#define GETKEY(dtt, tplp)   (*((tplp) +  (dtt)->keypos))
#define GETKEY_WITH_POS(Keypos, Tplp) (*((Tplp) + Keypos))


/* 
** Debugging
*/
#ifdef HARDDEBUG
#define DBG(args) /* XXXQQQ debug_printf args */
#ifdef	QQQ_TODO
static JudyDbTerm *traverse_until(JudyDbTerm *t, int *current, int to);
static void check_slot_pos(DbTableJudy *tb);
static void check_saved_stack(DbTableJudy *tb);
static int check_table_judy(JudyDbTerm *t);
#endif	/* QQQ_TODO */

#define JUDY_DEBUG
#else	/* HARDDEBUG */
#define DBG(args)
#endif	/* HARDDEBUG */

#ifdef JUDY_DEBUG
/*
** Debugging dump
*/

#ifdef	QQQ_TODO
static void do_dump_judy2(CIO fd, int show, JudyDbTerm *t, int offset);
#endif	/* QQQ_TODO */

#else
#endif

/*
 * Size calculations
 */
#define SIZ_OVERHEAD ((sizeof(JudyDbTerm)/sizeof(Eterm)) - 1)
#define SIZ_DBTERM(TDT) (SIZ_OVERHEAD + (TDT)->dbterm.size)

/*
** Datatypes
*/

/* 
 * This structure is filled in by analyze_pattern() for the select 
 * functions.
 */
struct mp_info {
    int all_objects;		/* True if complete objects are always
				 * returned from the match_spec (can use 
				 * copy_shallow on the return value) */
    int something_can_match;	/* The match_spec is not "impossible" */
    int some_limitation;	/* There is some limitation on the search
				 * area, i. e. least and/or most is set.*/
    int got_partial;		/* The limitation has a partially bound
				 * key */
    Eterm least;		/* The lowest matching key (possibly 
				 * partially bound expression) */
    Eterm most;                 /* The highest matching key (possibly 
				 * partially bound expression) */

    JudyDbTerm *save_term;      /* If the key is completely bound, this
				 * will be the Judy node we're searching
				 * for, otherwise it will be useless */
    Binary *mp;                 /* The compiled match program */
};

/*
 * Used by doit_select(_chunk)
 */
struct select_context {
    Process *p;
    Eterm accum;
    Binary *mp;
    Eterm end_condition;
    Eterm *lastobj;
    Sint32 max;
    int keypos;
    int all_objects;
    Sint got;
    Sint chunk_size;
};

/*
 * Used by doit_select_count
 */
struct select_count_context {
    Process *p;
    Binary *mp;
    Eterm end_condition;
    Eterm *lastobj;
    Sint32 max;
    int keypos;
    int all_objects;
    Sint got;
};

/*
 * Used by doit_select_delete
 */
struct select_delete_context {
    Process *p;
    DbTableJudy *tb;
    Uint accum;
    Binary *mp;
    Eterm end_condition;
    int erase_lastterm;
    JudyDbTerm *lastterm;
    Sint32 max;
    int keypos;
};

/*
** Forward declarations 
*/
static int debug_printf(char *, ...);
static int encode_key_term(DbTableJudy *, Eterm);
static int encode_key_string(DbTableJudy *, char *);
static int encode_key_bytes(DbTableJudy *, char *, int);
static int do_encode7(unsigned char *src, int srclen, unsigned char *dst, int dstmax);
#ifdef	COMMENT_OUT
static int do_decode7(unsigned char *src, int srclen, unsigned char *dst, int dstmax);
#endif	/* COMMENT_OUT */
static void do_free_judy(Pvoid_t root, char *);
static JudyDbTerm* get_term(JudyDbTerm* old, 
			    Eterm obj);
static int realloc_counter(JudyDbTerm** bp, Uint sz, 
			   Eterm new_counter, int counterpos);
static void free_judy_term(JudyDbTerm* p);
#ifdef	QQQ_TODO
static JudyDbTerm *linkout_judy(DbTableJudy *tb, Eterm key);
static JudyDbTerm *linkout_object_judy(DbTableJudy *tb, 
				       Eterm object);
static int balance_left(JudyDbTerm **this); 
static int balance_right(JudyDbTerm **this); 
static int delsub(JudyDbTerm **this); 
static JudyDbTerm *slot_search(Process *p, DbTableJudy *tb, Sint slot);
static JudyDbTerm *find_node(DbTableJudy *tb, Eterm key);
static JudyDbTerm **find_node2(DbTableJudy *tb, Eterm key);
static JudyDbTerm *find_next(DbTableJudy *tb, Eterm key);
static JudyDbTerm *find_prev(DbTableJudy *tb, Eterm key);
static JudyDbTerm *find_next_from_pb_key(DbTableJudy *tb, Eterm key);
static JudyDbTerm *find_prev_from_pb_key(DbTableJudy *tb, Eterm key);
static void traverse_backwards(DbTableJudy *tb,
			       Eterm lastkey,
			       int (*doit)(JudyDbTerm *, void *, int),
			       void *context); 
static void traverse_forward(DbTableJudy *tb,
			     Eterm lastkey,
			     int (*doit)(JudyDbTerm *, void *, int),
			     void *context); 
static int key_given(DbTableJudy *tb, Eterm pattern, JudyDbTerm **ret,
		     Eterm *partly_bound_key);
static int cmp_partly_bound(Eterm partly_bound_key, Eterm bound_key);
static int do_cmp_partly_bound(Eterm a, Eterm b, int *done);

static int analyze_pattern(DbTableJudy *tb, Eterm pattern, 
			   struct mp_info *mpi);
static int doit_select(JudyDbTerm *this, void *ptr, int forward);
static int doit_select_count(JudyDbTerm *this, void *ptr, int forward);
static int doit_select_chunk(JudyDbTerm *this, void *ptr, int forward);
static int doit_select_delete(JudyDbTerm *this, void *ptr, int forward);
static void do_dump_judy(CIO fd, int show, JudyDbTerm *t);

static int partly_bound_can_match_lesser(Eterm partly_bound_1, 
					 Eterm partly_bound_2);
static int partly_bound_can_match_greater(Eterm partly_bound_1, 
					  Eterm partly_bound_2); 
static int do_partly_bound_can_match_lesser(Eterm a, Eterm b, 
					    int *done);
static int do_partly_bound_can_match_greater(Eterm a, Eterm b, 
					     int *done);
static BIF_RETTYPE ets_select_reverse(Process *p, Eterm a1, 
				      Eterm a2, Eterm a3);
#endif	/* QQQ_TODO */

/*
** Static variables
*/

Export ets_select_reverse_exp;

/*
** External interface 
*/
void db_initialize_judy(void)
{
    /* QQQ Nothing to do here? */
};

/*
** Table interface routines ie what's called by the bif's 
*/

int db_create_judy(Process *p, DbTableJudy *tb)
{
    DBG(("%s: top", __FUNCTION__));
    tb->root = NULL;
    tb->memory = sizeof(DbTableJudy) / sizeof(Eterm);
    tb->binkey = safe_alloc_from(54, DB_JUDY_BINKEY_MAXSIZE);
    tb->enckey = safe_alloc_from(54, DB_JUDY_ENCKEY_MAXSIZE);
    return DB_ERROR_NONE;
}

/* release all memory occupied by a single table */
void free_judy_table(DbTableJudy *tb)
{
    DBG(("%s: top", __FUNCTION__));
    do_free_judy(tb->root, tb->enckey);
    if (tb->binkey)
	sys_free(tb->binkey);
    if (tb->enckey)
	sys_free(tb->enckey);
}

static void free_judy_term(JudyDbTerm* p)
{
    /* XXX More complexity is required here for bag types!! */
    db_free_term_data(&(p->dbterm));
    sys_free(p);
}

static void do_free_judy(Pvoid_t root, char *enckeybuf) 
{
    Word_t *judyval;
    int rc;
    struct timeval start, stop;	/* XXX */

    DBG(("%s: top", __FUNCTION__));
    if (root == NULL) {
	return;
    }
    enckeybuf[0] = '\0';		/* Start at top of table */
    JSLF(judyval, root, enckeybuf);
    if (judyval == NULL) {
	return;
    }
    do {
	free_judy_term((JudyDbTerm *) *judyval);
	/* Leave this node alone, let JSLFA() delete the whole Judy tree. */
	JSLN(judyval, root, enckeybuf);
    } while (judyval != NULL);
    gettimeofday(&start, NULL);	/* XXX */
    JSLFA(rc, root);
    gettimeofday(&stop, NULL);	/* XXX */
    debug_printf("XXX %s: JSLFA() start %u.%06u stop %u.%06u", __FUNCTION__, start.tv_sec, start.tv_usec, stop.tv_sec, stop.tv_usec); /* XXX */
}

int db_put_judy(Process *proc, DbTableJudy *tb, 
		Eterm obj, Eterm *ret)
{
    Eterm key;
    JudyDbTerm *newterm, *curterm;
    PWord_t judyval;

    DBG(("%s: top", __FUNCTION__));
    key = GETKEY(tb, tuple_val(obj));
    if (! encode_key_term(tb, key)) {
	return DB_ERROR_BADKEY;
    }
    JSLG(judyval, tb->root, tb->enckey);
    if (judyval != NULL && ((curterm = (JudyDbTerm *) *judyval) != NULL)) {
	/* XXX handle bag, duplicate bag, & ordered_set handling here */
	DBG(("%s: key exists, hardcoded 'set' behavior!", __FUNCTION__));
	tb->memory -= curterm->dbterm.size;
	newterm = get_term(curterm, obj);
	tb->memory += newterm->dbterm.size;
    } else {
	newterm = get_term(NULL, obj);
	tb->memory += SIZ_DBTERM(newterm);
	JSLI(judyval, tb->root, tb->enckey); /* XXX error handling! */
    }
    *judyval = (Word_t) newterm;
    *ret = am_true;    
    return DB_ERROR_NONE;
}

static JudyDbTerm*
get_term(JudyDbTerm* old, Eterm obj) 
{
    JudyDbTerm* p = db_get_term((old != NULL) ? &(old->dbterm) : NULL, 
				((char *) &(old->dbterm)) - ((char *) old),
				obj);
    return p;
}

int db_get_judy(Process *p, DbTableJudy *tb, 
		Eterm key, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm;
    Eterm copy;
    Eterm *hp;

    DBG(("%s: top", __FUNCTION__));
    if (! encode_key_term(tb, key)) {
	return DB_ERROR_BADKEY;
    }
    /*
     * XXX This is always a set, so we know exactly how large
     * the data is when we have found it.
     * The list created around it is purely for interface conformance.
     */
    
    JSLG(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	*ret = NIL;
    } else {
	hp = HAlloc(p, curterm->dbterm.size + 2);
	copy = copy_shallow(curterm->dbterm.v, 
			    curterm->dbterm.size, 
			    &hp, 
			    &MSO(p));
	*ret = CONS(hp, copy, NIL);
    }
    return DB_ERROR_NONE;
}

int db_member_judy(Process *p, DbTableJudy *tb, 
		   Eterm key, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm;

    DBG(("%s: top", __FUNCTION__));
    if (! encode_key_term(tb, key)) {
	return DB_ERROR_BADKEY;
    }
    JSLG(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	*ret = am_false;
    } else {
	*ret = am_true;
    }
    return DB_ERROR_NONE;
}

int db_first_judy(Process *p, DbTableJudy *tb, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm;
    Eterm e;
    Eterm *hp;
    Uint sz;

    if (! encode_key_string(tb, "")) {
	return DB_ERROR_UNSPEC;
    }
    JSLF(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }
    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, curterm->dbterm.tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
}

int db_next_judy(Process *p, DbTableJudy *tb, 
		 Eterm key, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm;
    Eterm e;
    Eterm *hp;
    Uint sz;

    if (! encode_key_term(tb, key)) {
	return DB_ERROR_BADKEY;
    }
    JSLN(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }
    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, curterm->dbterm.tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
}

int db_last_judy(Process *p, DbTableJudy *tb, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm;
    Eterm e;
    Eterm *hp;
    Uint sz;

    sys_memset(tb->binkey, 0xff, DB_JUDY_BINKEY_MAXSIZE);
    if (! encode_key_bytes(tb, tb->binkey, DB_JUDY_BINKEY_MAXSIZE)) {
	return DB_ERROR_UNSPEC;
    }
    JSLL(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }
    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, curterm->dbterm.tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
}

int db_prev_judy(Process *p, DbTableJudy *tb, 
		 Eterm key, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm;
    Eterm e;
    Eterm *hp;
    Uint sz;

    if (! encode_key_term(tb, key)) {
	return DB_ERROR_BADKEY;
    }
    JSLP(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }
    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, curterm->dbterm.tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
}

int db_erase_judy(Process *p, DbTableJudy *tb, 
		  Eterm key, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm;
    int rc;

    DBG(("%s: top", __FUNCTION__));
    *ret = am_true;
    if (! encode_key_term(tb, key)) {
	return DB_ERROR_BADKEY;
    }
    JSLG(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	/* do nothing */
    } else {
	free_judy_term(curterm);
	tb->memory -= curterm->dbterm.size;
	JSLD(rc, tb->root, tb->enckey);	/* XXX check error here & elsewhere?? */
    }
    return DB_ERROR_NONE;
}

/*
** XXX Boy, a person sure could have a lot of fun trying to optimize
** db_update_counter_METHOD().  If the typical operation were only
** updating the counter, then it would be cheaper (especially for
** bignum counters) to keep the latest counter value in the
** DbTableJudy struct, then update the tuple when an ets:lookup()
** is finally called ... except that any other query (e.g. ets:select())
** would have to have extra code added to handle this case.
**
** ... and to what end?  Is it *really* that important to have
** blindingly-fast write-mostly ETS counters?
*/

int db_update_counter_judy(Process *p, DbTableJudy *tb, 
			   Eterm key, Eterm incr, int counterpos, Eterm *ret)
{
    PWord_t judyval;
    JudyDbTerm *curterm, *old_curterm;
    int res;

    DBG(("%s: top", __FUNCTION__));
    if (! encode_key_term(tb, key)) {
	return DB_ERROR_BADKEY;
    }
    /* NOTE: Caller assures us that this is always a set */
    JSLG(judyval, tb->root, tb->enckey);
    if (judyval == NULL || ((curterm = (JudyDbTerm *) *judyval) == NULL)) {
	return DB_ERROR_BADKEY;
    } else {
	if (counterpos <= 0)
	    counterpos = tb->keypos + 1;
	old_curterm = curterm;
	res = db_do_update_counter(p, (void *) &curterm, curterm->dbterm.tpl,
				   counterpos, 
				   (int (*)(void *, Uint, Eterm, int))
				   &realloc_counter, incr, ret);
	if (res == DB_ERROR_NONE && curterm != old_curterm) {
	    *judyval = (Word_t) curterm;
	}
	return res;
    }
}

/*
 * Callback function for db_do_update_counter
 */
static int realloc_counter(JudyDbTerm** bp, Uint sz, 
			   Eterm new_counter, int counterpos)
{
    JudyDbTerm* b = *bp;
    return db_realloc_counter((void **) bp, &(b->dbterm),
			      ((char *) &(b->dbterm)) - ((char *) b),
			      sz, new_counter, counterpos);
}

#ifdef	QQQ_TODO

int db_get_element_tree(Process *p, DbTableTree *tb, 
			Eterm key, int ndex, Eterm *ret)
{
    /*
     * Look the node up:
     */
    Eterm *hp;
    TreeDbTerm *this;

    /*
     * This is always a set, so we know exactly how large
     * the data is when we have found it.
     * No list is created around elements in set's so there are no list
     * around the element here either.
     */
    
    this = find_node(tb,key);
    if (this == NULL) {
	return DB_ERROR_BADKEY;
    } else {
	Eterm element;
	Uint sz;
	if (ndex > arityval(this->dbterm.tpl[0])) {
	    return DB_ERROR_BADPARAM;
	}
	element = this->dbterm.tpl[ndex];
	sz = size_object(element);
	hp = HAlloc(p, sz);
	*ret = copy_struct(element, 
			   sz, 
			   &hp, 
			   &MSO(p));
    }
    return DB_ERROR_NONE;
}

int db_erase_object_tree(Process *p, DbTableTree *tb, 
			 Eterm object, Eterm *ret)
{
    TreeDbTerm *res;

    *ret = am_true;

    if ((res = linkout_object_tree(tb, object)) != NULL) {
	free_judy_term(res);
    }
    return DB_ERROR_NONE;
}


int db_slot_tree(Process *p, DbTableTree *tb, 
		 Eterm slot_term, Eterm *ret)
{
    Sint slot;
    TreeDbTerm *st;
    Eterm *hp;
    Eterm copy;

    /*
     * The notion of a "slot" is not natural in a tree, but we try to
     * simulate it by giving the n'th node in the tree instead.
     * Traversing a tree in this way is not very convenient, but by
     * using the saved stack we at least sometimes will get acceptable 
     * performance.
     */

    if (is_not_small(slot_term) ||
	((slot = signed_val(slot_term)) < 0) ||
	(slot > tb->nitems))
	return DB_ERROR_BADPARAM;

    if (slot == tb->nitems) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }

    /* 
     * We use the slot position and search from there, slot positions 
     * are counted from 1 and up.
     */
    ++slot;
    st = slot_search(p, tb, slot); 
    if (st == NULL) {
	*ret = am_false;
	return DB_ERROR_UNSPEC;
    }
    hp = HAlloc(p, st->dbterm.size + 2);
    copy = copy_shallow(st->dbterm.v, 
			st->dbterm.size, 
			&hp, 
			&MSO(p));
    *ret = CONS(hp, copy, NIL);
    return DB_ERROR_NONE;
}



static BIF_RETTYPE ets_select_reverse(Process *p, Eterm a1, Eterm a2, Eterm a3)
{
    Eterm list;
    Eterm result;
    Eterm* hp;
    Eterm* hend;

    int max_iter = CONTEXT_REDS * 10;

    if (is_nil(a1)) {
	hp = HAlloc(p, 3);
	BIF_RET(TUPLE2(hp,a2,a3));
    } else if (is_not_list(a1)) {
    error:
	BIF_ERROR(p, BADARG);
    }
    
    list = a1;
    result = a2;
    hp = hend = NULL;
    while (is_list(list)) {
	Eterm* pair = list_val(list);
	if (--max_iter == 0) {
	    BUMP_ALL_REDS(p);
	    HRelease(p, hp);
	    BIF_TRAP3(&ets_select_reverse_exp, p, list, result, a3);
	}
	if (hp == hend) {
	    hp = HAlloc(p, 64);
	    hend = hp + 64;
	}
	result = CONS(hp, CAR(pair), result);
	hp += 2;
	list = CDR(pair);
    }
    if (is_not_nil(list))  {
	goto error;
    }
    HRelease(p, hp);
    BUMP_REDS(p,CONTEXT_REDS - max_iter / 10);
    hp = HAlloc(p,3);
    BIF_RET(TUPLE2(hp, result, a3));
}

static BIF_RETTYPE bif_trap1(Export *bif,
			     Process *p, 
			     Eterm p1) 
{
    BIF_TRAP1(bif, p, p1);
}
    
static BIF_RETTYPE bif_trap3(Export *bif,
			     Process *p, 
			     Eterm p1, 
			     Eterm p2,
			     Eterm p3) 
{
    BIF_TRAP3(bif, p, p1, p2, p3);
}
    
/*
** This is called either when the select bif traps or when ets:select/1 
** is called. It does mostly the same as db_select_tree and may in either case
** trap to itself again (via the ets:select/1 bif).
** Note that this is common for db_select_tree and db_select_chunk_tree.
*/
int db_select_tree_continue(Process *p, 
			    DbTableTree *tb,
			    Eterm continuation,
			    Eterm *ret)
{
    struct select_context sc;
    unsigned sz;
    Eterm *hp; 
    Eterm lastkey;
    Eterm end_condition; 
    Binary *mp;
    Eterm key;
    Eterm *tptr;
    Sint chunk_size;
    int reverse;


#define RET_TO_BIF(Term, State) do { *ret = (Term); return State; } while(0);

    /* Decode continuation. We know it's a tuple but not the arity or 
       anything else */

    tptr = tuple_val(continuation);

    if (arityval(*tptr) != 8)
	RET_TO_BIF(NIL,DB_ERROR_BADPARAM);
    
    if (!is_small(tptr[4]) || !is_binary(tptr[5]) || 
	!(is_list(tptr[6]) || tptr[6] == NIL) || !is_small(tptr[7]) ||
	!is_small(tptr[8]))
	RET_TO_BIF(NIL,DB_ERROR_BADPARAM);
    
    lastkey = tptr[2];
    end_condition = tptr[3];
    mp = ((ProcBin *) binary_val(tptr[5]))->val;
    if (!(mp->flags & BIN_FLAG_MATCH_PROG))
	RET_TO_BIF(NIL,DB_ERROR_BADPARAM);
    chunk_size = signed_val(tptr[4]);

    sc.p = p;
    sc.accum = tptr[6];
    sc.mp = mp;
    sc.end_condition = NIL;
    sc.lastobj = NULL;
    sc.max = 1000;
    sc.keypos = tb->keypos;
    sc.all_objects = mp->flags & BIN_FLAG_ALL_OBJECTS;
    sc.chunk_size = chunk_size;
    reverse = unsigned_val(tptr[7]);
    sc.got = signed_val(tptr[8]);


    if (chunk_size) {
	if (reverse) {
	    traverse_backwards(tb, lastkey, &doit_select_chunk, &sc); 
	} else {
	    traverse_forward(tb, lastkey, &doit_select_chunk, &sc); 
	}
    } else {
	if (reverse) {
	    traverse_forward(tb, lastkey, &doit_select, &sc);
	} else {
	    traverse_backwards(tb, lastkey, &doit_select, &sc);
	}
    }

    BUMP_REDS(p, 1000 - sc.max);

    if (sc.max > 0) {
	if (chunk_size) {
	    Eterm *hp; 
	    unsigned sz;

	    if (sc.got < chunk_size || sc.lastobj == NULL) { 
		/* end of table, sc.lastobj may be NULL as we may have been
		   at the very last object in the table when trapping. */
		if (!sc.got) {
		    RET_TO_BIF(am_EOT, DB_ERROR_NONE);
		} else {
		    RET_TO_BIF(bif_trap3(&ets_select_reverse_exp, p,
					 sc.accum, NIL, am_EOT), 
			       DB_ERROR_NONE);
		}
	    }

	    key = GETKEY(tb, sc.lastobj);

	    sz = size_object(key);
	    hp = HAlloc(p, 9 + sz);
	    key = copy_struct(key, sz, &hp, &MSO(p));
	    continuation = TUPLE8
		(hp,
		 tptr[1],
		 key,
		 tptr[3], 
		 tptr[4],
		 tptr[5],
		 NIL,
		 tptr[7],
		 make_small(0));
	    RET_TO_BIF(bif_trap3(&ets_select_reverse_exp, p,
				 sc.accum, NIL, continuation), 
		       DB_ERROR_NONE);
	} else {
	    RET_TO_BIF(sc.accum, DB_ERROR_NONE);
	}
    }	
    key = GETKEY(tb, sc.lastobj);
    if (chunk_size) {
	if (end_condition != NIL && 
	    ((!reverse && cmp_partly_bound(end_condition,key) < 0) ||
	     (reverse && cmp_partly_bound(end_condition,key) > 0))) { 
	    /* done anyway */
	    if (!sc.got) {
		RET_TO_BIF(am_EOT, DB_ERROR_NONE);
	    } else {
		RET_TO_BIF(bif_trap3(&ets_select_reverse_exp, p, 
				     sc.accum, NIL, am_EOT), 
			   DB_ERROR_NONE);
	    }
	}
    } else {
	if (end_condition != NIL && 
	    ((!reverse && cmp_partly_bound(end_condition,key) > 0) ||
	     (reverse && cmp_partly_bound(end_condition,key) < 0))) { 
	    /* done anyway */
	    RET_TO_BIF(sc.accum,DB_ERROR_NONE);
	}
    }
    /* Not done yet, let's trap. */
    sz = size_object(key);
    hp = HAlloc(p, 9 + sz);
    key = copy_struct(key, sz, &hp, &MSO(p));
    continuation = TUPLE8
	(hp,
	 tptr[1],
	 key,
	 tptr[3], 
	 tptr[4],
	 tptr[5],
	 sc.accum,
	 tptr[7],
	 make_small(sc.got));
    RET_TO_BIF(bif_trap1(bif_export[BIF_ets_select_1], p, continuation), 
	       DB_ERROR_NONE);

#undef RET_TO_BIF
}


int db_select_tree(Process *p, DbTableTree *tb, 
		   Eterm pattern, int reverse, Eterm *ret)
{
    struct select_context sc;
    struct mp_info mpi;
    Eterm lastkey = NIL;
    Eterm key;
    Eterm continuation;
    unsigned sz;
    Eterm *hp; 
    TreeDbTerm *this;
    int errcode;
    Eterm mpb;


#define RET_TO_BIF(Term,RetVal) do { 	       	\
	if (mpi.mp != NULL) {			\
	    erts_match_set_free(mpi.mp);       	\
	}					\
	*ret = (Term); 				\
	return RetVal; 			        \
    } while(0)

    mpi.mp = NULL;

    sc.accum = NIL;
    sc.lastobj = NULL;
    sc.p = p;
    sc.max = 1000; 
    sc.end_condition = NIL;
    sc.keypos = tb->keypos;
    sc.got = 0;
    sc.chunk_size = 0;

    if ((errcode = analyze_pattern(tb, pattern, &mpi)) != DB_ERROR_NONE) {
	RET_TO_BIF(NIL,errcode);
    }

    if (!mpi.something_can_match) {
	RET_TO_BIF(NIL,DB_ERROR_NONE);  
	/* can't possibly match anything */
    }

    sc.mp = mpi.mp;
    sc.all_objects = mpi.all_objects;

    if (!mpi.got_partial && mpi.some_limitation && 
	cmp(mpi.least,mpi.most) == 0) {
	doit_select(mpi.save_term,&sc,0 /* direction doesn't matter */);
	RET_TO_BIF(sc.accum,DB_ERROR_NONE);
    }

    if (reverse) {
	if (mpi.some_limitation) {
	    if ((this = find_prev_from_pb_key(tb, mpi.least)) != NULL) {
		lastkey = GETKEY(tb, this->dbterm.tpl);
	    }
	    sc.end_condition = mpi.most;
	}
	
	traverse_forward(tb, lastkey, &doit_select, &sc);
    } else {
	if (mpi.some_limitation) {
	    if ((this = find_next_from_pb_key(tb, mpi.most)) != NULL) {
		lastkey = GETKEY(tb, this->dbterm.tpl);
	    }
	    sc.end_condition = mpi.least;
	}
	
	traverse_backwards(tb, lastkey, &doit_select, &sc);
    }
#ifdef HARDDEBUG
	erl_printf(CERR,"Least: ");
	display(mpi.least,CERR);
	erl_printf(CERR,"\r\nMost: ");
	display(mpi.most,CERR);
	erl_printf(CERR,"\r\n");
#endif
    BUMP_REDS(p, 1000 - sc.max);
    if (sc.max > 0) {
	RET_TO_BIF(sc.accum,DB_ERROR_NONE);
    }

    key = GETKEY(tb, sc.lastobj);
    sz = size_object(key);
    hp = HAlloc(p, 9 + sz + PROC_BIN_SIZE);
    key = copy_struct(key, sz, &hp, &MSO(p));
    if (mpi.all_objects)
	(mpi.mp)->flags |= BIN_FLAG_ALL_OBJECTS;
    mpb=db_make_mp_binary(p,mpi.mp,&hp);
	    
    continuation = TUPLE8
	(hp,
	 tb->id,
	 key,
	 sc.end_condition, /* From the match program, needn't be copied */
	 make_small(0), /* Chunk size of zero means not chunked to the
			   continuation BIF */
	 mpb,
	 sc.accum,
	 make_small(reverse),
	 make_small(sc.got));

    /* Don't free mpi.mp, so don't use macro */
    *ret = bif_trap1(bif_export[BIF_ets_select_1], p, continuation); 
    return DB_ERROR_NONE;

#undef RET_TO_BIF

}

    
/*
** This is called either when the select_count bif traps.
*/
int db_select_count_tree_continue(Process *p, 
				  DbTableTree *tb,
				  Eterm continuation,
				  Eterm *ret)
{
    struct select_count_context sc;
    unsigned sz;
    Eterm *hp; 
    Eterm lastkey;
    Eterm end_condition; 
    Binary *mp;
    Eterm key;
    Eterm *tptr;
    Eterm egot;


#define RET_TO_BIF(Term, State) do { *ret = (Term); return State; } while(0);

    /* Decode continuation. We know it's a tuple and everything else as
     this is only called by ourselves */

    /* continuation: 
       {Table, Lastkey, EndCondition, MatchProgBin, HowManyGot}*/

    tptr = tuple_val(continuation);

    if (arityval(*tptr) != 5)
	erl_exit(1,"Internal error in ets:select_count/1");
    
    lastkey = tptr[2];
    end_condition = tptr[3];
    mp = ((ProcBin *) binary_val(tptr[4]))->val;
    if (!(mp->flags & BIN_FLAG_MATCH_PROG))
	RET_TO_BIF(NIL,DB_ERROR_BADPARAM);

    sc.p = p;
    sc.mp = mp;
    sc.end_condition = NIL;
    sc.lastobj = NULL;
    sc.max = 1000;
    sc.keypos = tb->keypos;
    if (is_big(tptr[5])) {
	sc.got = big_to_uint32(tptr[5]);
    } else {
	sc.got = unsigned_val(tptr[5]);
    }

    traverse_backwards(tb, lastkey, &doit_select_count, &sc);

    BUMP_REDS(p, 1000 - sc.max);

    if (sc.max > 0) {
	RET_TO_BIF(make_small_or_big(sc.got,p), DB_ERROR_NONE);
    }	
    key = GETKEY(tb, sc.lastobj);
    if (end_condition != NIL && 
	(cmp_partly_bound(end_condition,key) > 0)) { 
	/* done anyway */
	RET_TO_BIF(make_small(sc.got),DB_ERROR_NONE);
    }
    /* Not done yet, let's trap. */
    egot = make_small_or_big(sc.got,p);
    sz = size_object(key);
    hp = HAlloc(p, 6 + sz);
    key = copy_struct(key, sz, &hp, &MSO(p));
    continuation = TUPLE5
	(hp,
	 tptr[1],
	 key,
	 tptr[3], 
	 tptr[4],
	 egot);
    RET_TO_BIF(bif_trap1(&ets_select_count_continue_exp, p, continuation), 
	       DB_ERROR_NONE);

#undef RET_TO_BIF
}


int db_select_count_tree(Process *p, DbTableTree *tb, 
			 Eterm pattern, Eterm *ret)
{
    struct select_count_context sc;
    struct mp_info mpi;
    Eterm lastkey = NIL;
    Eterm key;
    Eterm continuation;
    unsigned sz;
    Eterm *hp; 
    TreeDbTerm *this;
    int errcode;
    Eterm egot;
    Eterm mpb;


#define RET_TO_BIF(Term,RetVal) do { 	       	\
	if (mpi.mp != NULL) {			\
	    erts_match_set_free(mpi.mp);       	\
	}					\
	*ret = (Term); 				\
	return RetVal; 			        \
    } while(0)

    mpi.mp = NULL;

    sc.lastobj = NULL;
    sc.p = p;
    sc.max = 1000; 
    sc.end_condition = NIL;
    sc.keypos = tb->keypos;
    sc.got = 0;

    if ((errcode = analyze_pattern(tb, pattern, &mpi)) != DB_ERROR_NONE) {
	RET_TO_BIF(NIL,errcode);
    }

    if (!mpi.something_can_match) {
	RET_TO_BIF(make_small(0),DB_ERROR_NONE);  
	/* can't possibly match anything */
    }

    sc.mp = mpi.mp;
    sc.all_objects = mpi.all_objects;

    if (!mpi.got_partial && mpi.some_limitation && 
	cmp(mpi.least,mpi.most) == 0) {
	doit_select_count(mpi.save_term,&sc,0 /* dummy */);
	RET_TO_BIF(make_small_or_big(sc.got,p),DB_ERROR_NONE);
    }

    if (mpi.some_limitation) {
	if ((this = find_next_from_pb_key(tb, mpi.most)) != NULL) {
	    lastkey = GETKEY(tb, this->dbterm.tpl);
	}
	sc.end_condition = mpi.least;
    }
    
    traverse_backwards(tb, lastkey, &doit_select_count, &sc);
    BUMP_REDS(p, 1000 - sc.max);
    if (sc.max > 0) {
	RET_TO_BIF(make_small_or_big(sc.got,p),DB_ERROR_NONE);
    }

    egot = make_small_or_big(sc.got,p);
    key = GETKEY(tb, sc.lastobj);
    sz = size_object(key);
    hp = HAlloc(p, 6 + sz + PROC_BIN_SIZE);
    key = copy_struct(key, sz, &hp, &MSO(p));
    if (mpi.all_objects)
	(mpi.mp)->flags |= BIN_FLAG_ALL_OBJECTS;
    mpb = db_make_mp_binary(p,mpi.mp,&hp);
	    
    continuation = TUPLE5
	(hp,
	 tb->id,
	 key,
	 sc.end_condition, /* From the match program, needn't be copied */
	 mpb,
	 egot);

    /* Don't free mpi.mp, so don't use macro */
    *ret = bif_trap1(&ets_select_count_continue_exp, p, continuation); 
    return DB_ERROR_NONE;

#undef RET_TO_BIF

}

int db_select_chunk_tree(Process *p, DbTableTree *tb, 
			 Eterm pattern, Sint chunk_size,
			 int reverse,
			 Eterm *ret)
{
    struct select_context sc;
    struct mp_info mpi;
    Eterm lastkey = NIL;
    Eterm key;
    Eterm continuation;
    unsigned sz;
    Eterm *hp; 
    TreeDbTerm *this;
    int errcode;
    Eterm mpb;


#define RET_TO_BIF(Term,RetVal) do { 		\
	if (mpi.mp != NULL) {			\
	    erts_match_set_free(mpi.mp);	\
	}					\
	*ret = (Term); 				\
	return RetVal; 			        \
    } while(0)

    mpi.mp = NULL;

    sc.accum = NIL;
    sc.lastobj = NULL;
    sc.p = p;
    sc.max = 1000; 
    sc.end_condition = NIL;
    sc.keypos = tb->keypos;
    sc.got = 0;
    sc.chunk_size = chunk_size;

    if ((errcode = analyze_pattern(tb, pattern, &mpi)) != DB_ERROR_NONE) {
	RET_TO_BIF(NIL,errcode);
    }

    if (!mpi.something_can_match) {
	RET_TO_BIF(am_EOT,DB_ERROR_NONE);
	/* can't possibly match anything */
    }

    sc.mp = mpi.mp;
    sc.all_objects = mpi.all_objects;

    if (!mpi.got_partial && mpi.some_limitation && 
	cmp(mpi.least,mpi.most) == 0) {
	doit_select(mpi.save_term,&sc, 0 /* direction doesn't matter */);
	if (sc.accum != NIL) {
	    hp=HAlloc(p, 3);
	    RET_TO_BIF(TUPLE2(hp,sc.accum,am_EOT),DB_ERROR_NONE);
	} else {
	    RET_TO_BIF(am_EOT,DB_ERROR_NONE);
	}
    }

    if (reverse) {
	if (mpi.some_limitation) {
	    if ((this = find_next_from_pb_key(tb, mpi.most)) != NULL) {
		lastkey = GETKEY(tb, this->dbterm.tpl);
	    }
	    sc.end_condition = mpi.least;
	}

	traverse_backwards(tb, lastkey, &doit_select_chunk, &sc);
    } else {
	if (mpi.some_limitation) {
	    if ((this = find_prev_from_pb_key(tb, mpi.least)) != NULL) {
		lastkey = GETKEY(tb, this->dbterm.tpl);
	    }
	    sc.end_condition = mpi.most;
	}

	traverse_forward(tb, lastkey, &doit_select_chunk, &sc);
    }

    BUMP_REDS(p, 1000 - sc.max);
    if (sc.max > 0) {
	Eterm *hp; 
	unsigned sz;

	if (sc.got < chunk_size ||
	    sc.lastobj == NULL) { 
	    /* We haven't got all and we haven't trapped 
	       which should mean we are at the end of the 
	       table, sc.lastobj may be NULL if the table was empty */
	    
	    if (!sc.got) {
		RET_TO_BIF(am_EOT, DB_ERROR_NONE);
	    } else {
		RET_TO_BIF(bif_trap3(&ets_select_reverse_exp, p,
				     sc.accum, NIL, am_EOT), 
			   DB_ERROR_NONE);
	    }
	}

	key = GETKEY(tb, sc.lastobj);
	sz = size_object(key);
	hp = HAlloc(p, 9 + sz + PROC_BIN_SIZE);
	key = copy_struct(key, sz, &hp, &MSO(p));
	if (mpi.all_objects)
	    (mpi.mp)->flags |= BIN_FLAG_ALL_OBJECTS;
	mpb = db_make_mp_binary(p,mpi.mp,&hp);
	
	continuation = TUPLE8
	    (hp,
	     tb->id,
	     key,
	     sc.end_condition, /* From the match program, 
				  needn't be copied */
	     make_small(chunk_size),
	     mpb,
	     NIL,
	     make_small(reverse),
	     make_small(0));
	/* Don't let RET_TO_BIF macro free mpi.mp*/
	*ret = bif_trap3(&ets_select_reverse_exp, p,
			 sc.accum, NIL, continuation);
	return DB_ERROR_NONE; 
    }

    key = GETKEY(tb, sc.lastobj);
    sz = size_object(key);
    hp = HAlloc(p, 9 + sz + PROC_BIN_SIZE);
    key = copy_struct(key, sz, &hp, &MSO(p));

    if (mpi.all_objects)
	(mpi.mp)->flags |= BIN_FLAG_ALL_OBJECTS;
    mpb = db_make_mp_binary(p,mpi.mp,&hp);    
    continuation = TUPLE8
	(hp,
	 tb->id,
	 key,
	 sc.end_condition, /* From the match program, needn't be copied */
	 make_small(chunk_size),
	 mpb,
	 sc.accum,
	 make_small(reverse),
	 make_small(sc.got));
    /* Don't let RET_TO_BIF macro free mpi.mp*/
    *ret = bif_trap1(bif_export[BIF_ets_select_1], p, continuation);
    return DB_ERROR_NONE;

#undef RET_TO_BIF

}

/*
** This is called when select_delete traps
*/
int db_select_delete_continue_tree(Process *p, 
				   DbTableTree *tb,
				   Eterm continuation,
				   Eterm *ret)
{
    struct select_delete_context sc;
    unsigned sz;
    Eterm *hp; 
    Eterm lastkey;
    Eterm end_condition; 
    Binary *mp;
    Eterm key;
    Eterm *tptr;


#define RET_TO_BIF(Term, State) do { 		\
	if (sc.erase_lastterm) {		\
	    free_judy_term(sc.lastterm);		\
	}					\
	*ret = (Term); 				\
	return State; 				\
    } while(0);

    /* Decode continuation. We know it's correct, this can only be called
       by trapping */

    tptr = tuple_val(continuation);

    lastkey = tptr[2];
    end_condition = tptr[3];

    sc.erase_lastterm = 0; /* Before first RET_TO_BIF */
    sc.lastterm = NULL;

    mp = ((ProcBin *) binary_val(tptr[4]))->val;
    sc.p = p;
    sc.tb = tb;
    if (is_big(tptr[5])) {
	sc.accum = big_to_uint32(tptr[5]);
    } else {
	sc.accum = unsigned_val(tptr[5]);
    }
    sc.mp = mp;
    sc.end_condition = NIL;
    sc.max = 1000;
    sc.keypos = tb->keypos;

    traverse_backwards(tb, lastkey, &doit_select_delete, &sc);

    BUMP_REDS(p, 1000 - sc.max);

    if (sc.max > 0) {
	RET_TO_BIF(make_small_or_big(sc.accum, p), DB_ERROR_NONE);
    }	
    key = GETKEY(tb, (sc.lastterm)->dbterm.tpl);
    if (end_condition != NIL && 
	cmp_partly_bound(end_condition,key) > 0) { /* done anyway */
	RET_TO_BIF(make_small_or_big(sc.accum,p),DB_ERROR_NONE);
    }
    /* Not done yet, let's trap. */
    sz = size_object(key);
    hp = HAlloc(p, 6 + sz);
    key = copy_struct(key, sz, &hp, &MSO(p));
    continuation = TUPLE5
	(hp,
	 tptr[1],
	 key,
	 tptr[3], 
	 tptr[4],
	 make_small_or_big(sc.accum,p));
    RET_TO_BIF(bif_trap1(&ets_select_delete_continue_exp, p, continuation), 
	       DB_ERROR_NONE);

#undef RET_TO_BIF
}

int db_select_delete_tree(Process *p, DbTableTree *tb, 
			  Eterm pattern, Eterm *ret)
{
    struct select_delete_context sc;
    struct mp_info mpi;
    Eterm lastkey = NIL;
    Eterm key;
    Eterm continuation;
    unsigned sz;
    Eterm *hp; 
    TreeDbTerm *this;
    int errcode;
    Eterm mpb;

#define RET_TO_BIF(Term,RetVal) do { 	       	\
	if (mpi.mp != NULL) {			\
	    erts_match_set_free(mpi.mp);       	\
	}					\
	if (sc.erase_lastterm) {                \
	    free_judy_term(sc.lastterm);             \
	}                                       \
	*ret = (Term); 				\
	return RetVal; 			        \
    } while(0)

    mpi.mp = NULL;

    sc.accum = 0;
    sc.erase_lastterm = 0;
    sc.lastterm = NULL;
    sc.p = p;
    sc.max = 1000; 
    sc.end_condition = NIL;
    sc.keypos = tb->keypos;
    sc.tb = tb;
    
    if ((errcode = analyze_pattern(tb, pattern, &mpi)) != DB_ERROR_NONE) {
	RET_TO_BIF(0,errcode);
    }

    if (!mpi.something_can_match) {
	RET_TO_BIF(make_small(0),DB_ERROR_NONE);  
	/* can't possibly match anything */
    }

    sc.mp = mpi.mp;

    if (!mpi.got_partial && mpi.some_limitation && 
	cmp(mpi.least,mpi.most) == 0) {
	doit_select_delete(mpi.save_term,&sc, 0 /* direction doesn't 
						   matter */);
	RET_TO_BIF(make_small_or_big(sc.accum,p),DB_ERROR_NONE);
    }

    if (mpi.some_limitation) {
	if ((this = find_next_from_pb_key(tb, mpi.most)) != NULL) {
	    lastkey = GETKEY(tb, this->dbterm.tpl);
	}
	sc.end_condition = mpi.least;
    }

    traverse_backwards(tb, lastkey, &doit_select_delete, &sc);
    BUMP_REDS(p, 1000 - sc.max);

    if (sc.max > 0) {
	RET_TO_BIF(make_small_or_big(sc.accum,p), DB_ERROR_NONE);
    }

    key = GETKEY(tb, (sc.lastterm)->dbterm.tpl);
    sz = size_object(key);
    hp = HAlloc(p, 6 + sz + PROC_BIN_SIZE);
    key = copy_struct(key, sz, &hp, &MSO(p));
    mpb = db_make_mp_binary(p,mpi.mp,&hp);
    
    continuation = TUPLE5
	(hp,
	 tb->id,
	 key,
	 sc.end_condition, /* From the match program, needn't be copied */
	 mpb,
	 make_small_or_big(sc.accum, p));

    /* Don't free mpi.mp, so don't use macro */
    if (sc.erase_lastterm) {
	free_judy_term(sc.lastterm);
    }
    *ret = bif_trap1(&ets_select_delete_continue_exp, p, continuation); 
    return DB_ERROR_NONE;

#undef RET_TO_BIF

}

/*
** Other interface routines (not directly coupled to one bif)
*/

/* Display hash table contents (for dump) */
void db_print_tree(CIO fd, 
		   int show,
		   DbTableTree *tb)
{
#ifdef JUDY_DEBUG
    if (show)
	erl_printf(fd, "\r\nTree data dump:\r\n"
		   "------------------------------------------------\r\n");
    do_dump_tree2(fd, show, tb->root, 0);
    if (show)
	erl_printf(fd, "\r\n"
		   "------------------------------------------------\r\n");
#else
    erl_printf(fd,"Ordered set (AVL tree), Elements: %d\n", tb->nitems);
    do_dump_tree(fd, show, tb->root);
#endif
}

static void do_db_tree_foreach_offheap(TreeDbTerm *,
				       void (*)(ErlOffHeap *, void *),
				       void *);

void
erts_db_tree_foreach_offheap(DbTableTree *tab,
			     void (*func)(ErlOffHeap *, void *),
			     void * arg)
{
    do_db_tree_foreach_offheap(tab->root, func, arg);
}


/*
** Functions for internal use
*/


static void
do_db_tree_foreach_offheap(TreeDbTerm *tdbt,
			   void (*func)(ErlOffHeap *, void *),
			   void * arg)
{
    if(!tdbt)
	return;
    do_db_tree_foreach_offheap(tdbt->left, func, arg);
    (*func)(&(tdbt->dbterm.off_heap), arg);
    do_db_tree_foreach_offheap(tdbt->right, func, arg);
}

static TreeDbTerm *linkout_tree(DbTableTree *tb, 
				Eterm key)
{
    TreeDbTerm **tstack[STACK_NEED];
    int tpos = 0;
    int dstack[STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    TreeDbTerm **this = &tb->root;
    int c;
    int dir;
    TreeDbTerm *q = NULL;

    /*
     * Somewhat complicated, deletion in an AVL tree,
     * The two helpers balance_left and balance_right are used to
     * keep the balance. As in insert, we do the stacking ourselves.
     */

    tb->stack_pos = tb->slot_pos = 0;
    dstack[dpos++] = DIR_END;
    for (;;) {
	if (!*this) { /* Failure */
	    return NULL;
	} else if ((c = cmp(key,GETKEY(tb,(*this)->dbterm.tpl))) < 0) { 
	    dstack[dpos++] = DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else if (c > 0) { /* go right */
	    dstack[dpos++] = DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	} else { /* Equal key, found the one to delete*/
	    q = (*this);
	    if (q->right == NULL) {
		(*this) = q->left;
		state = 1;
	    } else if (q->left == NULL) {
		(*this) = q->right;
		state = 1;
	    } else {
		dstack[dpos++] = DIR_LEFT;
		tstack[tpos++] = this;
		state = delsub(this);
	    }
	    --(tb->nitems);
	    tb->memory -= SIZ_DBTERM(q);
	    break;
	}
    }
    while (state && ( dir = dstack[--dpos] ) != DIR_END) {
	this = tstack[--tpos];
	if (dir == DIR_LEFT) {
	    state = balance_left(this);
	} else {
	    state = balance_right(this);
	}
    }
    return q;
}

static TreeDbTerm *linkout_object_tree(DbTableTree *tb, 
				       Eterm object)
{
    TreeDbTerm **tstack[STACK_NEED];
    int tpos = 0;
    int dstack[STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    TreeDbTerm **this = &tb->root;
    int c;
    int dir;
    TreeDbTerm *q = NULL;
    Eterm key;

    /*
     * Somewhat complicated, deletion in an AVL tree,
     * The two helpers balance_left and balance_right are used to
     * keep the balance. As in insert, we do the stacking ourselves.
     */

    
    key = GETKEY(tb, tuple_val(object));

    tb->stack_pos = tb->slot_pos = 0;
    dstack[dpos++] = DIR_END;
    for (;;) {
	if (!*this) { /* Failure */
	    return NULL;
	} else if ((c = cmp(key,GETKEY(tb,(*this)->dbterm.tpl))) < 0) { 
	    dstack[dpos++] = DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else if (c > 0) { /* go right */
	    dstack[dpos++] = DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	} else { /* Equal key, found the only possible matching object*/
	    if (!eq(object,make_tuple((*this)->dbterm.tpl))) {
		return NULL;
	    }
	    q = (*this);
	    if (q->right == NULL) {
		(*this) = q->left;
		state = 1;
	    } else if (q->left == NULL) {
		(*this) = q->right;
		state = 1;
	    } else {
		dstack[dpos++] = DIR_LEFT;
		tstack[tpos++] = this;
		state = delsub(this);
	    }
	    --(tb->nitems);
	    tb->memory -= SIZ_DBTERM(q);
	    break;
	}
    }
    while (state && ( dir = dstack[--dpos] ) != DIR_END) {
	this = tstack[--tpos];
	if (dir == DIR_LEFT) {
	    state = balance_left(this);
	} else {
	    state = balance_right(this);
	}
    }
    return q;
}

/*
** For the select functions, analyzes the pattern and determines which
** part of the tree should be searched. Also compiles the match program
*/
static int analyze_pattern(DbTableTree *tb, Eterm pattern, 
			   struct mp_info *mpi)
{
    Eterm lst, tpl, ttpl;
    Eterm *matches,*guards, *bodies;
    Eterm sbuff[30];
    Eterm *buff = sbuff;
    Eterm *ptpl;
    int i;
    int num_heads = 0;
    Eterm key;
    Eterm partly_bound;
    int res;
    Eterm least = 0;
    Eterm most = 0;
    TreeDbTerm *this;

    mpi->some_limitation = 1;
    mpi->got_partial = 0;
    mpi->something_can_match = 0;
    mpi->mp = NULL;
    mpi->all_objects = 1;
    mpi->save_term = NULL;

    for (lst = pattern; is_list(lst); lst = CDR(list_val(lst)))
	++num_heads;

    if (lst != NIL) {/* proper list... */
	return DB_ERROR_BADPARAM;
    }
    if (num_heads > 10) {
	buff = safe_alloc_from(41, sizeof(Eterm) * num_heads * 3);
    }

    matches = buff;
    guards = buff + num_heads;
    bodies = buff + (num_heads * 2);

    i = 0;
    for(lst = pattern; is_list(lst); lst = CDR(list_val(lst))) {
	Eterm body;
	ttpl = CAR(list_val(lst));
	if (!is_tuple(ttpl)) {
	    if (buff != sbuff) { 
		sys_free(buff);
	    }
	    return DB_ERROR_BADPARAM;
	}
	ptpl = tuple_val(ttpl);
	if (ptpl[0] != make_arityval(3U)) {
	    if (buff != sbuff) { 
		sys_free(buff);
	    }
	    return DB_ERROR_BADPARAM;
	}
	matches[i] = tpl = ptpl[1];
	guards[i] = ptpl[2];
	bodies[i] = body = ptpl[3];
	if (!is_list(body) || CDR(list_val(body)) != NIL ||
	    CAR(list_val(body)) != am_DollarUnderscore) {
	    mpi->all_objects = 0;
	}
	++i;

	partly_bound = NIL;
	res = key_given(tb, tpl, &this, &partly_bound);
	if ( res >= 0 ) {   /* Can match something */
	    key = 0;
	    mpi->something_can_match = 1;
	    if (res > 0) {
		mpi->save_term = this;
		key = GETKEY(tb,tuple_val(tpl)); 
	    } else if (partly_bound != NIL) {
		mpi->got_partial = 1;
		key = partly_bound;
	    } else {
		mpi->some_limitation = 0;
	    }
	    if (key != 0) {
		if (least == 0 || 
		    partly_bound_can_match_lesser
		    (key,least)) {
		    least = key;
		}
		if (most == 0 || 
		    partly_bound_can_match_greater
		    (key,most)) {
		    most = key;
		}
	    }
	}
    }
    mpi->least = least;
    mpi->most = most;
    if (mpi->something_can_match) {
	if ((mpi->mp = db_match_compile(matches, guards, bodies,
					num_heads, DCOMP_TABLE, NULL)) 
	    == NULL) {
	    if (buff != sbuff) { 
		sys_free(buff);
	    }
	    return DB_ERROR_BADPARAM;
	}
    }
    if (buff != sbuff) { 
	sys_free(buff);
    }
    return DB_ERROR_NONE;
}

static void do_dump_tree(CIO fd, int show, TreeDbTerm *t)
{
    if (t == NULL)
	return;
    do_dump_tree(fd, show, t->left);
    if (show) {
	display(make_tuple(t->dbterm.tpl), fd);
	erl_printf(fd, "\n");
    }
    do_dump_tree(fd, show, t->right); 
    return;
}

/*
 * Deletion helpers
 */
static int balance_left(TreeDbTerm **this) 
{
    TreeDbTerm *p, *p1, *p2;
    int b1, b2, h = 1;
    
    p = *this;
    switch (p->balance) {
    case -1:
	p->balance = 0;
	break;
    case 0:
	p->balance = 1;
	h = 0;
	break;
    case 1:
	p1 = p->right;
	b1 = p1->balance;
	if (b1 >= 0) { /* Single RR rotation */
	    p->right = p1->left;
	    p1->left = p;
	    if (b1 == 0) {
		p->balance = 1;
		p1->balance = -1;
		h = 0;
	    } else {
		p->balance = p1->balance = 0;
	    }
	    (*this) = p1;
	} else { /* Double RL rotation */
	    p2 = p1->left;
	    b2 = p2->balance;
	    p1->left = p2->right;
	    p2->right = p1;
	    p->right = p2->left;
	    p2->left = p;
	    p->balance = (b2 == 1) ? -1 : 0;
	    p1->balance = (b2 == -1) ? 1 : 0;
	    p2->balance = 0;
	    (*this) = p2;
	}
	break;
    }
    return h;
}

static int balance_right(TreeDbTerm **this) 
{
    TreeDbTerm *p, *p1, *p2;
    int b1, b2, h = 1;
    
    p = *this;
    switch (p->balance) {
    case 1:
	p->balance = 0;
	break;
    case 0:
	p->balance = -1;
	h = 0;
	break;
    case -1:
	p1 = p->left;
	b1 = p1->balance;
	if (b1 <= 0) { /* Single LL rotation */
	    p->left = p1->right;
	    p1->right = p;
	    if (b1 == 0) {
		p->balance = -1;
		p1->balance = 1;
		h = 0;
	    } else {
		p->balance = p1->balance = 0;
	    }
	    (*this) = p1;
	} else { /* Double LR rotation */
	    p2 = p1->right;
	    b2 = p2->balance;
	    p1->right = p2->left;
	    p2->left = p1;
	    p->left = p2->right;
	    p2->right = p;
	    p->balance = (b2 == -1) ? 1 : 0;
	    p1->balance = (b2 == 1) ? -1 : 0;
	    p2->balance = 0;
	    (*this) = p2;
	}
    }
    return h;
}

static int delsub(TreeDbTerm **this) 
{
    TreeDbTerm **tstack[STACK_NEED];
    int tpos = 0;
    TreeDbTerm *q = (*this);
    TreeDbTerm **r = &(q->left);
    int h;

    /*
     * Walk down the tree to the right and search 
     * for a void right child, pick that child out
     * and return it to be put in the deleted 
     * object's place.
     */
    
    while ((*r)->right != NULL) {
	tstack[tpos++] = r;
	r = &((*r)->right);
    }
    *this = *r;
    *r = (*r)->left;
    (*this)->left = q->left;
    (*this)->right = q->right;
    (*this)->balance = q->balance;
    tstack[0] = &((*this)->left);
    h = 1;
    while (tpos && h) {
	r = tstack[--tpos];
	h = balance_right(r);
    }
    return h;
}

/*
 * Helper for db_slot
 */

static TreeDbTerm *slot_search(Process *p, DbTableTree *tb, Sint slot)
{
    TreeDbTerm *this;
    TreeDbTerm *tmp;

    if (slot == 1) { /* Don't search from where we are if we are 
			looking for the first slot */
	tb->slot_pos = 0;
    }

    if (tb->slot_pos == 0) { /* clear stack if slot positions 
				are not recorded */
	tb->stack_pos = 0;
    }
    if (EMPTY_NODE(tb)) {
	this = tb->root;
	if (this == NULL)
	    return NULL;
	while (this->left != NULL){
	    PUSH_NODE(tb, this);
	    this = this->left;
	}
	PUSH_NODE(tb, this);
	tb->slot_pos = 1;
    }
    this = TOP_NODE(tb);
    while (tb->slot_pos != slot && this != NULL) {
	if (slot > tb->slot_pos) {
	    if (this->right != NULL) {
		this = this->right;
		while (this->left != NULL) {
		    PUSH_NODE(tb, this);
		    this = this->left;
		}
		PUSH_NODE(tb, this);
	    } else {
		for (;;) {
		    tmp = POP_NODE(tb);
		    this = TOP_NODE(tb);
		    if (this == NULL || this->left == tmp)
			break;
		}
	    }		
	    ++(tb->slot_pos);
	} else {
	    if (this->left != NULL) {
		this = this->left;
		while (this->right != NULL) {
		    PUSH_NODE(tb, this);
		    this = this->right;
		}
		PUSH_NODE(tb, this);
	    } else {
		for (;;) {
		    tmp = POP_NODE(tb);
		    this = TOP_NODE(tb);
		    if (this == NULL || this->right == tmp)
			break;
		}
	    }		
	    --(tb->slot_pos);
	}
    }
    return this;
}

/*
 * Find next and previous in sort order
 */

static TreeDbTerm *find_next(DbTableTree *tb, Eterm key)
{
    TreeDbTerm *this;
    TreeDbTerm *tmp;
    int c;

    if(( this = TOP_NODE(tb)) != NULL) {
	if (!CMP_EQ(GETKEY(tb, this->dbterm.tpl),key)) {
	    /* Start from the beginning */
	    tb->stack_pos = tb->slot_pos = 0;
	}
    }
    if (EMPTY_NODE(tb)) { /* Have to rebuild the stack */
	if (( this = tb->root ) == NULL)
	    return NULL;
	for (;;) {
	    PUSH_NODE(tb, this);
	    if (( c = cmp(GETKEY(tb, this->dbterm.tpl),key) ) < 0) {
		if (this->right == NULL) /* We are at the previos 
					    and the element does
					    not exist */
		    break;
		else
		    this = this->right;
	    } else if (c > 0) {
		if (this->left == NULL) /* Done */
		    return this;
		else
		    this = this->left;
	    } else
		break;
	}
    }
    /* The next element from this... */
    if (this->right != NULL) {
	this = this->right;
	PUSH_NODE(tb,this);
	while (this->left != NULL) {
	    this = this->left;
	    PUSH_NODE(tb, this);
	}
	if (tb->slot_pos > 0) 
	    ++(tb->slot_pos);
    } else {
	do {
	    tmp = POP_NODE(tb);
	    if (( this = TOP_NODE(tb)) == NULL) {
		tb->slot_pos = 0;
		return NULL;
	    }
	} while (this->right == tmp);
	if (tb->slot_pos > 0) 
	    ++(tb->slot_pos);
    }
    return this;
}

static TreeDbTerm *find_prev(DbTableTree *tb, Eterm key)
{
    TreeDbTerm *this;
    TreeDbTerm *tmp;
    int c;

    if(( this = TOP_NODE(tb)) != NULL) {
	if (!CMP_EQ(GETKEY(tb, this->dbterm.tpl),key)) {
	    /* Start from the beginning */
	    tb->stack_pos = tb->slot_pos = 0;
	}
    }
    if (EMPTY_NODE(tb)) { /* Have to rebuild the stack */
	if (( this = tb->root ) == NULL)
	    return NULL;
	for (;;) {
	    PUSH_NODE(tb, this);
	    if (( c = cmp(GETKEY(tb, this->dbterm.tpl),key) ) > 0) {
		if (this->left == NULL) /* We are at the next 
					   and the element does
					   not exist */
		    break;
		else
		    this = this->left;
	    } else if (c < 0) {
		if (this->right == NULL) /* Done */
		    return this;
		else
		    this = this->right;
	    } else
		break;
	}
    }
    /* The previous element from this... */
    if (this->left != NULL) {
	this = this->left;
	PUSH_NODE(tb,this);
	while (this->right != NULL) {
	    this = this->right;
	    PUSH_NODE(tb, this);
	}
	if (tb->slot_pos > 0) 
	    --(tb->slot_pos);
    } else {
	do {
	    tmp = POP_NODE(tb);
	    if (( this = TOP_NODE(tb)) == NULL) {
		tb->slot_pos = 0;
		return NULL;
	    }
	} while (this->left == tmp);
	if (tb->slot_pos > 0) 
	    --(tb->slot_pos);
    }
    return this;
}

static TreeDbTerm *find_next_from_pb_key(DbTableTree *tb, Eterm key)
{
    TreeDbTerm *this;
    TreeDbTerm *tmp;
    int c;

    /* spool the stack, we have to "re-search" */
    tb->stack_pos = tb->slot_pos = 0;
    if (( this = tb->root ) == NULL)
	return NULL;
    for (;;) {
	PUSH_NODE(tb, this);
	if (( c = cmp_partly_bound(key,GETKEY(tb, this->dbterm.tpl)) ) >= 0) {
	    if (this->right == NULL) {
		do {
		    tmp = POP_NODE(tb);
		    if (( this = TOP_NODE(tb)) == NULL) {
			return NULL;
		    }
		} while (this->right == tmp);
		return this;
	    } else
		this = this->right;
	} else /*if (c < 0)*/ {
	    if (this->left == NULL) /* Done */
		return this;
	    else
		this = this->left;
	} 
    }
}

static TreeDbTerm *find_prev_from_pb_key(DbTableTree *tb, Eterm key)
{
    TreeDbTerm *this;
    TreeDbTerm *tmp;
    int c;

    /* spool the stack, we have to "re-search" */
    tb->stack_pos = tb->slot_pos = 0;
    if (( this = tb->root ) == NULL)
	return NULL;
    for (;;) {
	PUSH_NODE(tb, this);
	if (( c = cmp_partly_bound(key,GETKEY(tb, this->dbterm.tpl)) ) <= 0) {
	    if (this->left == NULL) {
		do {
		    tmp = POP_NODE(tb);
		    if (( this = TOP_NODE(tb)) == NULL) {
			return NULL;
		    }
		} while (this->left == tmp);
		return this;
	    } else
		this = this->left;
	} else /*if (c < 0)*/ {
	    if (this->right == NULL) /* Done */
		return this;
	    else
		this = this->right;
	} 
    }
}


/*
 * Just lookup a node
 */
static TreeDbTerm *find_node(DbTableTree *tb, Eterm key)
{
    TreeDbTerm *this;
    int res;

    if(!EMPTY_NODE(tb) && 
       CMP_EQ(GETKEY(tb, ( this = TOP_NODE(tb) )->dbterm.tpl), key))
	return this;
    this = tb->root;
    while (this != NULL && 
	   ( res = cmp(key, GETKEY(tb, this->dbterm.tpl)) ) != 0) {
	if (res < 0)
	    this = this->left;
	else
	    this = this->right;
    }
    return this;
}

/*
 * Lookup a node and return the address of the node pointer in the tree
 */
static TreeDbTerm **find_node2(DbTableTree *tb, Eterm key)
{
    TreeDbTerm **this;
    int res;

    this = &tb->root;
    while ((*this) != NULL && 
	   ( res = cmp(key, GETKEY(tb, (*this)->dbterm.tpl)) ) != 0) {
	if (res < 0)
	    this = &((*this)->left);
	else
	    this = &((*this)->right);
    }
    if (*this == NULL)
	return NULL;
    return this;
}

/*
 * Traverse the tree with a callback function, used by db_match_xxx
 */
static void traverse_backwards(DbTableTree *tb,
			       Eterm lastkey,
			       int (*doit)(TreeDbTerm *, void *, int),
			       void *context) 
{
    TreeDbTerm *this, *next;

    if (lastkey == NIL) {
	tb->stack_pos = tb->slot_pos = 0;
	if (( this = tb->root ) == NULL) {
	    return;
	}
	while (this != NULL) {
	    PUSH_NODE(tb, this);
	    this = this->right;
	}
	this = TOP_NODE(tb);
	next = find_prev(tb, GETKEY(tb, this->dbterm.tpl));
	if (!((*doit)(this, context, 0)))
	    return;
    } else {
	next = find_prev(tb, lastkey);
    }

    while ((this = next) != NULL) {
	next = find_prev(tb, GETKEY(tb, this->dbterm.tpl));
	if (!((*doit)(this, context, 0)))
	    return;
    }
}

/*
 * Traverse the tree with a callback function, used by db_match_xxx
 */
static void traverse_forward(DbTableTree *tb,
			     Eterm lastkey,
			     int (*doit)(TreeDbTerm *, void *, int),
			     void *context) 
{
    TreeDbTerm *this, *next;

    if (lastkey == NIL) {
	tb->stack_pos = tb->slot_pos = 0;
	if (( this = tb->root ) == NULL) {
	    return;
	}
	while (this != NULL) {
	    PUSH_NODE(tb, this);
	    this = this->left;
	}
	this = TOP_NODE(tb);
	next = find_next(tb, GETKEY(tb, this->dbterm.tpl));
	if (!((*doit)(this, context, 1)))
	    return;
    } else {
	next = find_next(tb, lastkey);
    }

    while ((this = next) != NULL) {
	next = find_next(tb, GETKEY(tb, this->dbterm.tpl));
	if (!((*doit)(this, context, 1)))
	    return;
    }
}

/*
 * Returns 0 if not given 1 if given and -1 on no possible match
 * if key is given and ret != NULL, *ret is set to point 
 * to the object concerned;
 */
static int key_given(DbTableTree *tb, Eterm pattern, TreeDbTerm **ret, 
		     Eterm *partly_bound)
{
    TreeDbTerm *this;
    Eterm key;

    if (pattern == am_Underscore || db_is_variable(pattern) != -1)
	return 0;
    key = db_getkey(tb->keypos, pattern);
    if (is_non_value(key))
	return -1;  /* can't possibly match anything */
    if (!db_has_variable(key)) {   /* Bound key */
	if(( this = find_node(tb, key) ) == NULL)
	    return -1;
	if (ret != NULL) 
	    *ret = this; 
	return 1;
    } else if (partly_bound != NULL && key != am_Underscore && 
	       db_is_variable(key) < 0)
	*partly_bound = key;
	
    return 0;
}



static int do_cmp_partly_bound(Eterm a, Eterm b, int *done)
{
    Eterm* aa;
    Eterm* bb;
    Eterm a_hdr;
    Eterm b_hdr;
    int i;
    int j;

    /* A variable matches anything */
    if (is_atom(a) && (a == am_Underscore || (db_is_variable(a) >= 0))) {
	*done = 1;
	return 0;
    }
    if (a == b)
	return 0;
    
    switch (a & _TAG_PRIMARY_MASK) {
    case TAG_PRIMARY_LIST:
	if (!is_list(b)) {
	    return cmp(a,b);
	}
	aa = list_val(a);
	bb = list_val(b);
	while (1) {
	    if ((j = do_cmp_partly_bound(*aa++, *bb++, done)) != 0 || *done) 
		return j;
	    if (*aa==*bb)
		return 0;
	    if (is_not_list(*aa) || is_not_list(*bb))
		return do_cmp_partly_bound(*aa, *bb, done);
	    aa = list_val(*aa);
	    bb = list_val(*bb);
	}
    case TAG_PRIMARY_BOXED:
	if ((b & _TAG_PRIMARY_MASK) != TAG_PRIMARY_BOXED) {
	    return cmp(a,b);
	}
	a_hdr = ((*boxed_val(a)) & _TAG_HEADER_MASK) >> _TAG_PRIMARY_SIZE;
	b_hdr = ((*boxed_val(a)) & _TAG_HEADER_MASK) >> _TAG_PRIMARY_SIZE;
	if (a_hdr != b_hdr) {
	    return cmp(a, b);
	}
	if (a_hdr == (_TAG_HEADER_ARITYVAL >> _TAG_PRIMARY_SIZE)) {
	    aa = tuple_val(a);
	    bb = tuple_val(b);
	    /* compare the arities */
	    i = arityval(*aa);	/* get the arity*/
	    if (i < arityval(*bb)) return(-1);
	    if (i > arityval(*bb)) return(1);
	    while (i--) {
		if ((j = do_cmp_partly_bound(*++aa, *++bb, done)) != 0 
		    || *done) 
		    return j;
	    }
	    return 0;
	}
	/* Drop through */
      default:
	  return cmp(a, b);
    }
}

static int cmp_partly_bound(Eterm partly_bound_key, Eterm bound_key) 
{
    int done = 0;
    int ret = do_cmp_partly_bound(partly_bound_key, bound_key, &done);
#ifdef HARDDEBUG
    erl_printf(CERR,"\r\ncmp_partly_bound: ");
    display(partly_bound_key,CERR);
    if (ret < 0)
	erl_printf(CERR," < ");
    else if (ret > 0)
	erl_printf(CERR," > ");
    else
	erl_printf(CERR," == ");
    display(bound_key,CERR);
    erl_printf(CERR,"\r\n");
#endif
    return ret;
}

/*
** For partly_bound debugging....
**
BIF_RETTYPE ets_testnisse_2(BIF_ALIST_2)
BIF_ADECL_2
{
    Eterm r1 = make_small(partly_bound_can_match_lesser(BIF_ARG_1,
							BIF_ARG_2));
    Eterm r2 = make_small(partly_bound_can_match_greater(BIF_ARG_1,
							 BIF_ARG_2));
    Eterm *hp = HAlloc(BIF_P,3);
    Eterm ret;

    ret = TUPLE2(hp,r1,r2);
    BIF_RET(ret);
}
**
*/
static int partly_bound_can_match_lesser(Eterm partly_bound_1, 
					 Eterm partly_bound_2) 
{
    int done = 0;
    int ret = do_partly_bound_can_match_lesser(partly_bound_1, 
					       partly_bound_2, 
					       &done);
#ifdef HARDDEBUG
    erl_printf(CERR,"\r\npartly_bound_can_match_lesser: ");
    display(partly_bound_1,CERR);
    if (ret)
	erl_printf(CERR," can match lesser than ");
    else
	erl_printf(CERR," can not match lesser than ");
    display(partly_bound_2,CERR);
    erl_printf(CERR,"\r\n");
#endif
    return ret;
}

static int partly_bound_can_match_greater(Eterm partly_bound_1, 
					  Eterm partly_bound_2) 
{
    int done = 0;
    int ret = do_partly_bound_can_match_greater(partly_bound_1, 
						partly_bound_2, 
						&done);
#ifdef HARDDEBUG
    erl_printf(CERR,"\r\npartly_bound_can_match_greater: ");
    display(partly_bound_1,CERR);
    if (ret)
	erl_printf(CERR," can match greater than ");
    else
	erl_printf(CERR," can not match greater than ");
    display(partly_bound_2,CERR);
    erl_printf(CERR,"\r\n");
#endif
    return ret;
}

static int do_partly_bound_can_match_lesser(Eterm a, Eterm b, 
					    int *done)
{
    Eterm* aa;
    Eterm* bb;
    int i;
    int j;

    if (is_atom(a) && (a == am_Underscore || 
		       (db_is_variable(a) >= 0))) {
	*done = 1;
	if (is_atom(b) && (b == am_Underscore || 
			   (db_is_variable(b) >= 0))) {
	    return 0;
	} else {
	    return 1;
	}
    } else if (is_atom(b) && (b == am_Underscore || 
			      (db_is_variable(b) >= 0))) {
	*done = 1;
	return 0;
    }

    if (a == b)
	return 0;

    if (not_eq_tags(a,b)) {
	*done = 1;
	return (cmp(a, b) < 0) ? 1 : 0;
    }

    /* we now know that tags are the same */
    switch (tag_val_def(a)) {
    case TUPLE_DEF:
	aa = tuple_val(a);
	bb = tuple_val(b);
	/* compare the arities */
	if (arityval(*aa) < arityval(*bb)) return 1;
	if (arityval(*aa) > arityval(*bb)) return 0;
	i = arityval(*aa);	/* get the arity*/
	while (i--) {
	    if ((j = do_partly_bound_can_match_lesser(*++aa, *++bb, 
						      done)) != 0 
		|| *done) 
		return j;
	}
	return 0;
    case LIST_DEF:
	aa = list_val(a);
	bb = list_val(b);
	while (1) {
	    if ((j = do_partly_bound_can_match_lesser(*aa++, *bb++, 
						      done)) != 0 
		|| *done) 
		return j;
	    if (*aa==*bb)
		return 0;
	    if (is_not_list(*aa) || is_not_list(*bb))
		return do_partly_bound_can_match_lesser(*aa, *bb, 
							done);
	    aa = list_val(*aa);
	    bb = list_val(*bb);
	}
    default:
	if((i = cmp(a, b)) != 0) {
	    *done = 1;
	}
	return (i < 0) ? 1 : 0;
    }
}

static int do_partly_bound_can_match_greater(Eterm a, Eterm b, 
					    int *done)
{
    Eterm* aa;
    Eterm* bb;
    int i;
    int j;

    if (is_atom(a) && (a == am_Underscore || 
		       (db_is_variable(a) >= 0))) {
	*done = 1;
	if (is_atom(b) && (b == am_Underscore || 
			   (db_is_variable(b) >= 0))) {
	    return 0;
	} else {
	    return 1;
	}
    } else if (is_atom(b) && (b == am_Underscore || 
			      (db_is_variable(b) >= 0))) {
	*done = 1;
	return 0;
    }

    if (a == b)
	return 0;

    if (not_eq_tags(a,b)) {
	*done = 1;
	return (cmp(a, b) > 0) ? 1 : 0;
    }

    /* we now know that tags are the same */
    switch (tag_val_def(a)) {
    case TUPLE_DEF:
	aa = tuple_val(a);
	bb = tuple_val(b);
	/* compare the arities */
	if (arityval(*aa) < arityval(*bb)) return 0;
	if (arityval(*aa) > arityval(*bb)) return 1;
	i = arityval(*aa);	/* get the arity*/
	while (i--) {
	    if ((j = do_partly_bound_can_match_greater(*++aa, *++bb, 
						      done)) != 0 
		|| *done) 
		return j;
	}
	return 0;
    case LIST_DEF:
	aa = list_val(a);
	bb = list_val(b);
	while (1) {
	    if ((j = do_partly_bound_can_match_greater(*aa++, *bb++, 
						      done)) != 0 
		|| *done) 
		return j;
	    if (*aa==*bb)
		return 0;
	    if (is_not_list(*aa) || is_not_list(*bb))
		return do_partly_bound_can_match_greater(*aa, *bb, 
							done);
	    aa = list_val(*aa);
	    bb = list_val(*bb);
	}
    default:
	if((i = cmp(a, b)) != 0) {
	    *done = 1;
	}
	return (i > 0) ? 1 : 0;
    }
}

/*
 * Callback functions for the different match functions
 */

static int doit_select(TreeDbTerm *this, void *ptr, int forward)
{
    struct select_context *sc = (struct select_context *) ptr;
    Eterm ret;
    Uint32 dummy;

    sc->lastobj = this->dbterm.tpl;
    
    if (sc->end_condition != NIL && 
	((forward && 
	  cmp_partly_bound(sc->end_condition, 
			   GETKEY_WITH_POS(sc->keypos, 
					   this->dbterm.tpl)) < 0) ||
	 (!forward && 
	  cmp_partly_bound(sc->end_condition, 
			   GETKEY_WITH_POS(sc->keypos, 
					   this->dbterm.tpl)) > 0))) {
	return 0;
    }
    ret = db_prog_match(sc->p, sc->mp,
			make_tuple(this->dbterm.tpl), 
			0, &dummy);
    if (is_value(ret)) {
	Uint sz;
	Eterm *hp;
	if (sc->all_objects) {
	    hp = HAlloc(sc->p, this->dbterm.size + 2);
	    ret = copy_shallow(this->dbterm.v,
				     this->dbterm.size,
				     &hp,
			             &MSO(sc->p));
	} else {
	    sz = size_object(ret);
	    hp = HAlloc(sc->p, sz + 2);
	    ret = copy_struct(ret, sz, 
			      &hp, &MSO(sc->p));
	}
	sc->accum = CONS(hp, ret, sc->accum);
    }
    if (--(sc->max) <= 0) {
	return 0;
    }
    return 1;
}

static int doit_select_count(TreeDbTerm *this, void *ptr, int forward)
{
    struct select_count_context *sc = (struct select_count_context *) ptr;
    Eterm ret;
    Uint32 dummy;

    sc->lastobj = this->dbterm.tpl;
    
    /* Always backwards traversing */
    if (sc->end_condition != NIL && 
	(cmp_partly_bound(sc->end_condition, 
			  GETKEY_WITH_POS(sc->keypos, 
					  this->dbterm.tpl)) > 0)) {
	return 0;
    }
    ret = db_prog_match(sc->p, sc->mp,
			make_tuple(this->dbterm.tpl), 
			0, &dummy);
    if (ret == am_true) {
	++(sc->got);
    }
    if (--(sc->max) <= 0) {
	return 0;
    }
    return 1;
}

static int doit_select_chunk(TreeDbTerm *this, void *ptr, int forward)
{
    struct select_context *sc = (struct select_context *) ptr;
    Eterm ret;
    Uint32 dummy;

    sc->lastobj = this->dbterm.tpl;
    
    if (sc->end_condition != NIL && 
	((forward && 
	  cmp_partly_bound(sc->end_condition, 
			   GETKEY_WITH_POS(sc->keypos, 
					   this->dbterm.tpl)) < 0) ||
	 (!forward && 
	  cmp_partly_bound(sc->end_condition, 
			   GETKEY_WITH_POS(sc->keypos, 
					   this->dbterm.tpl)) > 0))) {
	return 0;
    }

    ret = db_prog_match(sc->p, sc->mp,
			make_tuple(this->dbterm.tpl), 
			0, &dummy);
    if (is_value(ret)) {
	Uint sz;
	Eterm *hp;

	++(sc->got);
	if (sc->all_objects) {
	    hp = HAlloc(sc->p, this->dbterm.size + 2);
	    ret = copy_shallow(this->dbterm.v,
				     this->dbterm.size,
				     &hp,
			             &MSO(sc->p));
	} else {
	    sz = size_object(ret);
	    hp = HAlloc(sc->p, sz + 2);
	    ret = copy_struct(ret, sz, &hp, &MSO(sc->p));
	}
	sc->accum = CONS(hp, ret, sc->accum);
    }
    if (--(sc->max) <= 0 || sc->got == sc->chunk_size) {
	return 0;
    }
    return 1;
}


static int doit_select_delete(TreeDbTerm *this, void *ptr, int forward)
{
    struct select_delete_context *sc = (struct select_delete_context *) ptr;
    Eterm ret;
    Uint32 dummy;
    Eterm key;

    if (sc->erase_lastterm)
	free_judy_term(sc->lastterm);
    sc->erase_lastterm = 0;
    sc->lastterm = this;
    
    if (sc->end_condition != NIL && 
	cmp_partly_bound(sc->end_condition, 
			 GETKEY_WITH_POS(sc->keypos, 
					 this->dbterm.tpl)) > 0)
	return 0;
    ret = db_prog_match(sc->p, sc->mp,
			make_tuple(this->dbterm.tpl), 
			0, &dummy);
    if (ret == am_true) {
	key = GETKEY(sc->tb, this->dbterm.tpl);
	linkout_tree(sc->tb, key);
	sc->erase_lastterm = 1;
	++sc->accum;
    }
    if (--(sc->max) <= 0) {
	return 0;
    }
    return 1;
}

#ifdef JUDY_DEBUG
static void do_dump_tree2(CIO fd, int show, TreeDbTerm *t, int offset)
{
    if (t == NULL)
	return 0;
    do_dump_tree2(fd, show, t->right, offset + 4);
    if (show) {
	int i;
	for (i = 0; i < offset; ++i)
	    erl_printf(fd, " ");
	display(make_tuple(t->dbterm.tpl), fd);
	erl_printf(fd, "(addr = 0x%08X, bal = %d)\r\n", t, t->balance);
    }
    do_dump_tree2(fd, show, t->left, offset + 4); 
    return sum;
}

#endif

#endif	/* QQQ_TODO */

/*
** This 7-bits-in-each-8-bits encoding scheme is a hack to get around
** the limitations of the JudySL design assumptions.  However, with a
** couple of modifications to the JudySL library, some improvements
** could be made:
**
** 1. Have the MSB of the last byte of a string be zero, eliminating
** the need for a trailing 0 byte.
**
** 2. Change the encoding to a 29-bits-in-each-32-bits scheme, where
** the top three MSBs (too bad 2 bits isn't feasible!?) signify how
** many bytes (or fractions thereof) are stored in this particular
** word:
**
** xxx122byte2233byte3344byte445by5 <- worst-case map of where bytes bits
** --------------------------------
** 33222222222211111111110000000000
** 10987654321098765432109876543210
**
** Hrm, 2 signalling bits might be worth the extra complexity, but 3?
** Gack.
*/

#ifdef	XXX_DELETEME
main(int argc, char *argv[])
{
    unsigned long	start, stop;
    unsigned char	cur[MAXLEN], last[MAXLEN], unfmt[MAXLEN];
    int			bytes, lencur, lenlast, lenunfmt;
    unsigned long	i, val;
    unsigned char	*p;
    int			res;

    if (argc != 3) {
	fprintf(stderr, "usage: %s start-num top-num\n", argv[0]);
	exit(1);
    }
    start = atoi(argv[1]);
    stop = atoi(argv[2]);
    for (i = start; i <= stop; i++) {
	val = ntohl(i);
	if (i < (1 << 8)) {
	    p = ((unsigned char *) &val) + 3;
	    bytes = 1;
	} else if (i < (1 << 16)) {
	    p = ((unsigned char *) &val) + 2;
	    bytes = 2;
	} else if (i < (1 << 24)) {
	    p = ((unsigned char *) &val) + 1;
	    bytes = 3;
	} else {
	    p = ((unsigned char *) &val);
	    bytes = 4;
	}
	lencur = do_encode7(p, bytes, cur, MAXLEN);
	if (i != start && lencur == lenlast) {
	    if ((res = memcmp(last, cur, lencur)) >= 0) {
		fprintf(stderr, "Error at i = %lu res = %d, exiting\n", i, res);
		exit(2);
	    }
	}
	/* Unformat to verify reverse operation */
	lenunfmt = do_decode7(cur, lencur, unfmt, MAXLEN);
	if ((lenunfmt != bytes || memcmp(p, unfmt, bytes)) != 0) {
	    fprintf(stderr, "Unformat error at i = %lu, exiting\n", i);
	    exit(3);
	}
	/* Prep for next iteration */
	memcpy(last, cur, lencur);
	lenlast = lencur;
    }

    do_encode7("1234567890", 11, cur, 64);

    exit(0);
}
#endif	/* XXX_DELETEME */

/*
** encode_key_term() -- Return 1 if 7-bits-in-8 encoding of a term
**	is successful, 0 otherwise.
*/

static int
encode_key_term(DbTableJudy *tb, Eterm key)
{
    int size;
    byte *endp;

    DBG(("%s: top", __FUNCTION__));
    size = encode_size_struct(key, TERM_TO_BINARY_DFLAGS);
    if (size > DB_JUDY_BINKEY_MAXSIZE) {
	return 0;
    }
    if ((endp = enc_term(NULL, key, tb->binkey, TERM_TO_BINARY_DFLAGS)) == NULL) {
	DBG(("%s: enc_term failed", __FUNCTION__));
	return 0;
    }
    size = (char *) endp - tb->binkey;
    return encode_key_bytes(tb, tb->binkey, size);
}

static int
encode_key_string(DbTableJudy *tb, char *s)
{
    return encode_key_bytes(tb, s, strlen(s));
}

static int
encode_key_bytes(DbTableJudy *tb, char *s, int size)
{
    int encsize;

    if ((encsize = do_encode7(tb->binkey, size, tb->enckey, DB_JUDY_ENCKEY_MAXSIZE)) < 0) {
	DBG(("%s: do_encode7 failed", __FUNCTION__));
	return 0;
    }
    DBG(("%s: encsize = %d", __FUNCTION__, encsize));
    return 1;
}

static int
do_encode7(unsigned char *src, int srclen, unsigned char *dst, int dstmax)
{
    unsigned char	carry = 0;/* Bits left over from last src byte */
    int			carrylen = 0;
    int			remain;	/* Bits remaining in current encoded byte */
    unsigned char	tmp;
    unsigned char	curmask, carrymask;
    unsigned char	cur;
    int			dstlen = 0;

    if (srclen < 0 || dstlen > dstmax) {
	return -1;
    }
    while (srclen--) {
	tmp = 128;	/* Set high bit */
	remain = 7;
	if (carrylen > 0) {
	    /* Bits in carry are always "flush right" (in the LSBs) */
	    tmp |= (carry << (7 - carrylen));
	    remain -= carrylen;
	}
	if (remain == 0) {
	    *dst++ = tmp;
	    dstlen++;
	    if (dstlen > dstmax) {
		return -1;
	    }
	    tmp = 128;
	    remain = 7;
	}
	curmask = ~((1 << (8 - remain)) - 1);
	carrymask = ~curmask;
	cur = ((*src) & curmask) >> (8 - remain);
	tmp |= cur;
	carry = *src & carrymask;
	carrylen = 8 - remain;
	src++;
	*dst++ = tmp;
	dstlen++;
	if (dstlen > dstmax) {
	    return -1;
	}
    }
    /* There may be some straggler bits in carry, write them, too. */
    if (carrylen > 0) {
	tmp = 128;
	tmp |= (carry << (7 - carrylen));
	*dst++ = tmp;
	dstlen++;
	if (dstlen > dstmax) {
	    return -1;
	}
    }
    *dst = '\0';
    dstlen++;
    return dstlen;
}

#ifdef	COMMENT_OUT
static int
do_decode7(unsigned char *src, int srclen, unsigned char *dst, int dstmax)
{
    unsigned char	carry = 0;
    int			carryfilled = 0;
    int			fillbits;
    unsigned char	curmask, carrymask;
    unsigned char	cur;
    int			dstlen = 0;

    /* srclen is intentionally unused: these strings are NUL-terminated */
    if (dstlen > dstmax) {
	return -1;
    }
    while (*src != '\0') {
	if (carryfilled == 0) {
	    /* This is the easy one.... */
	    /* Bits in carry are always "flush right" (in the LSBs) */
	    carry = (*src) & 0x7f;
	    carryfilled = 7;
	} else {
	    fillbits = 8 - carryfilled; /* # src contributes after carry */
	    carrymask = (1 << (carryfilled - 1)) - 1;
	    curmask = 0x7f ^ carrymask;
	    cur  = carry << fillbits;
	    cur |= ((*src) & curmask) >> (7 - fillbits);
	    *dst++ = cur;
	    dstlen++;
	    if (dstlen > dstmax) {
		return -1;
	    }
	    carry = (*src) & carrymask;
	    carryfilled--;
	}
	src++;
    }
    return dstlen;
}
#endif	/* COMMENT_OUT */


static int
debug_printf(char *fmt, ...)
{
    va_list ap;
    int	rc;

    va_start(ap, fmt);
    rc = vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, "\r\n");
    return rc;
}

#ifdef HARDDEBUG

#ifdef	QQQ_TODO
void db_check_table_tree(DbTableTree *tb)
{
    check_table_tree(tb->root);
    check_saved_stack(tb);
    check_slot_pos(tb);
#endif	/* QQQ_TODO */

#ifdef	QQQ_TODO
static TreeDbTerm *traverse_until(TreeDbTerm *t, int *current, int to)
{
    TreeDbTerm *tmp;
    if (t == NULL) 
	return NULL;
    tmp = traverse_until(t->left, current, to);
    if (tmp != NULL)
	return tmp;
    ++(*current);
    if (*current == to)
	return t;
    return traverse_until(t->right, current, to);
}

static void check_slot_pos(DbTableTree *tb)
{
    int pos = 0;
    TreeDbTerm *t;
    if (tb->slot_pos == 0 || tb->stack_pos == 0)
	return;
    t = traverse_until(tb->root, &pos, tb->slot_pos);
    if (t != tb->stack[tb->stack_pos - 1]) {
	erl_printf(CERR, "Slot position does not correspont with stack, "
		   "element position %d is really 0x%08X, when stack says "
		   "it's 0x%08X\r\n", tb->slot_pos, t, 
		   tb->stack[tb->stack_pos - 1]);
	do_dump_tree2(CERR, 1, tb->root, 0);
    }
}
	

static void check_saved_stack(DbTableTree *tb)
{
     TreeDbTerm *t = tb->root;
     int n = 0;
     if (tb->stack_pos == 0)
	 return;
     if (t != tb->stack[n]) {
	 erl_printf(CERR,"tb->stack[0] is 0x%08X, should be 0x%08X\r\n",
		    tb->stack[0], t);
	 do_dump_tree2(CERR, 1, tb->root, 0);
	 return;
     }
     while (n < tb->stack_pos) {
	 if (t == NULL) {
	     erl_printf(CERR, "NULL pointer in tree when stack not empty,"
			" stack depth is %d\r\n", n);
	     do_dump_tree2(CERR, 1, tb->root, 0);
	     return;
	 }
	 n++;
	 if (n < tb->stack_pos) {
	     if (tb->stack[n] == t->left)
		 t = t->left;
	     else if (tb->stack[n] == t->right)
		 t = t->right;
	     else {
		 erl_printf(CERR, "tb->stack[%d] == 0x%08X does not "
			    "represent child pointer in tree!"
			    "(left == 0x%08X, right == 0x%08X\r\n", 
			    n, tb->stack[n], t->left, t->right);
		 do_dump_tree2(CERR, 1, tb->root, 0);
		 return;
	     }
	 }
     }
}

static int check_table_tree(TreeDbTerm *t)
{
    int lh, rh;
    if (t == NULL)
	return 0;
    lh = check_table_tree(t->left);
    rh = check_table_tree(t->right);
    if ((rh - lh) != t->balance) {
	erl_printf(CERR, "Invalid tree balance for this node:\r\n");
	erl_printf(CERR,"balance = %d, left = 0x%08X, right = 0x%08X\r\n"
		   "data =",
		   t->balance, t->left, t->right);
	display(make_tuple(t->dbterm.tpl),CERR);
	erl_printf(CERR,"\r\nDump:\r\n---------------------------------\r\n");
	do_dump_tree2(CERR, 1, t, 0);
	erl_printf(CERR,"\r\n---------------------------------\r\n");
    }
    return ((rh > lh) ? rh : lh) + 1;
}
#endif	/* QQQ_TODO */
	
#endif

/*
** XXX XXX XXX This stuff should go in a separate file, but I'm too
** lazy to do it now.
*/

/*
** btree.c -- From _Programs and Data Structures in C_, Leendert Ammeraal,
** John Wiley & Sons Ltd, 1987.  XXX License terms unclear.
**
** Portions by Scott Lystig Fritchie, 2003.
*/

#include <stdio.h>
#include <ctype.h>

/* Stack stuff borrowed from erl_db_tree.c */
#define STACK_NEED 50		/* More than we need for 2^32 elements */

#define PUSH_NODE(Dtt, Tdt, Idx)                   \
do {                                               \
    (Dtt)->stack[(Dtt)->stack_pos].node = (Tdt);   \
    (Dtt)->stack[(Dtt)->stack_pos].idx = (Idx);    \
    (Dtt)->stack_pos++;                            \
} while (0)

#define POP_NODE(Dtt, Node, Idx)                         \
do {                                                     \
    if ((Dtt)->stack_pos > 0) {                          \
        (Dtt)->stack_pos--;                              \
        (Node) = (Dtt)->stack[((Dtt)->stack_pos)].node;  \
        (Idx) = (Dtt)->stack[((Dtt)->stack_pos)].idx;    \
    } else {                                             \
        (Node) = NULL;                                   \
        (Idx) = -666;                                    \
    }                                                    \
} while (0)

#define TOP_NODE(Dtt, Node, Idx)                             \
do {                                                         \
    if ((Dtt)->stack_pos > 0) {                              \
        (Node) = (Dtt)->stack[((Dtt)->stack_pos) - 1].node;  \
        (Idx) =  (Dtt)->stack[((Dtt)->stack_pos) - 1].idx;   \
    } else {                                                 \
        (Node) = NULL;                                       \
        (Idx) = -666;                                        \
    }                                                        \
} while (0)

static void do_free_btree(Node_t *root);

/* Declarations for B-tree library internal stuff */
static Node_t *insert(DbTableBtree *, Eterm, Node_t *);
static Node_t *delete(DbTableBtree *, Eterm, Node_t *);
static DbTerm *search(DbTableBtree *tb, Eterm, Node_t *);
static int binsearch(DbTableBtree *tb, Eterm x, DbTerm **a, int n);
static void found(Node_t *t, int i);
static void notfound(Eterm x);
static int ins(DbTableBtree *, Eterm x, Node_t *t, DbTerm **y, Node_t **u);
static int del(DbTableBtree *, Eterm x, Node_t *t, Node_t *root);
static unsigned long print_tree(Node_t *, int, int, CIO);

/*
** External interface 
*/
void db_initialize_btree(void)
{
    /* QQQ Nothing to do here? */
};

int
db_create_btree(Process *p, DbTableBtree *tb)
{
    DBG(("%s: top", __FUNCTION__));
    tb->root = NULL;
    tb->memory = sizeof(DbTableBtree) / sizeof(Eterm);
    tb->stack = safe_alloc_from(54, sizeof(Btree_Stack_t) * STACK_NEED);
    tb->stack_pos = 0;
    return DB_ERROR_NONE;
}

/* release all memory occupied by a single table */
void
free_btree_table(DbTableBtree *tb)
{
    DBG(("%s: top", __FUNCTION__));
    do_free_btree(tb->root);
    if (tb->stack != NULL) {
	sys_free(tb->stack);
    }
}

static void
do_free_btree(Node_t *root)
{
    int	i;

    if (root == NULL) {
	return;
    }
    for (i = 0; i < root->cnt; i++) {
	do_free_btree(root->ptr[i]);
	db_free_term_data(root->key[i]);
	sys_free(root->key[i]);
    }
    /* Always one more pointer than there are keys */
    do_free_btree(root->ptr[root->cnt]);
    sys_free(root);
}

int
db_put_btree(Process *proc, DbTableBtree *tb, 
		 Eterm obj, Eterm *ret)
{
    DBG(("%s: top", __FUNCTION__));
    if ((tb->root = insert(tb, obj, tb->root)) == NULL) {
	fprintf(stderr, "XXX Should never happen!\r\n");
	abort();
    }
    *ret = am_true;    
    return DB_ERROR_NONE;
}

int
db_get_btree(Process *p, DbTableBtree *tb, 
		Eterm key, Eterm *ret)
{
    DbTerm *term;
    Eterm copy;
    Eterm *hp;

    DBG(("%s: top", __FUNCTION__));
    if ((term = search(tb, key, tb->root)) == NULL) {
	*ret = NIL;
    } else {
	/*
	 * XXX This is always a set, so we know exactly how large
	 * the data is when we have found it.
	 * The list created around it is purely for interface conformance.
	 */
	hp = HAlloc(p, term->size + 2);
	copy = copy_shallow(term->v, 
			    term->size, 
			    &hp, 
			    &MSO(p));
	*ret = CONS(hp, copy, NIL);
    }
    return DB_ERROR_NONE;
}

int
db_member_btree(Process *p, DbTableBtree *tb, 
		   Eterm key, Eterm *ret)
{
    DBG(("%s: top", __FUNCTION__));
    if (search(tb, key, tb->root) == NULL) {
	*ret = am_false;
    } else {
	*ret = am_true;
    }
    return DB_ERROR_NONE;
}

void
db_print_btree(CIO fd, int show, DbTableBtree *tb)
{
    unsigned long n;

    erl_printf(fd, "\r\n");
    n = print_tree(tb->root, 4, 0, fd);
    erl_printf(fd, "Total number of nodes: %lu\r\n", n);
}

int
db_first_btree(Process *p, DbTableBtree *tb, Eterm *ret)
{
    Node_t *t, *bottom;
    DbTerm *term;
    Eterm e;
    Eterm *hp;
    Uint sz;

    tb->stack_pos = 0;
    if (tb->root == NULL) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }
    for (t = tb->root; t != NULL; t = t->ptr[0]) {
	PUSH_NODE(tb, t, 0);
	bottom = t;
    }
    term = bottom->key[0];

    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, term->tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
}

int
db_last_btree(Process *p, DbTableBtree *tb, Eterm *ret)
{
    Node_t *t, *bottom;
    DbTerm *term;
    Eterm e;
    Eterm *hp;
    Uint sz;

    tb->stack_pos = 0;
    if (tb->root == NULL) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }
    for (t = tb->root; t != NULL; t = t->ptr[t->cnt]) {
	PUSH_NODE(tb, t, t->cnt - 1);
	bottom = t;
    }
    term = bottom->key[bottom->cnt - 1];

    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, term->tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
}

int
db_next_btree(Process *p, DbTableBtree *tb, Eterm key, Eterm *ret)
{
    *ret = am_EOT;
    return DB_ERROR_NONE;
#ifdef	XXX_INCOMPLETE
    Node_t *t;
    DbTerm *term = NULL;
    int idx;
    Eterm e;
    Eterm *hp;
    Uint sz;

    if (tb->root == NULL ||
	(t = search_next(tb, tb->root, key, &idx)) == NULL) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }

    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, t->key[idx]->tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
#endif	/* XXX_INCOMPLETE */
}

#ifdef	XXX_INCOMPLETE

static Node_t *search_next(DbTableBtree *tb, Node_t *t, Eterm key, int *idx);

static Node_t *
search_next(DbTableBtree *tb, Node_t *t, Eterm key, int *idx)
{
    int	i;
    int c;
    Eterm termkey;

    if (t == NULL) {
	return NULL;
    }
    i = binsearch(tb, key, t->key, t->cnt);
    if (i == t->cnt) {
	if (t->key[i] == NULL) {
	    return NULL;
	} else {
	    return search_next(tb, t->key[i], key, idx);
	}
    }
    if (CMP_LE(key, GETKEY(tb, t-> ZZZ left off here, is this method feasible??
}
#endif	/* XXX_INCOMPLETE */

#ifdef	XXX_HOSED
int
db_next_btree(Process *p, DbTableBtree *tb, Eterm key, Eterm *ret)
{
    DbTerm *term = NULL;
    int i = -9982;
    Node_t *t;
    Eterm e;
    Eterm *hp;
    Uint sz;
    int		n;
    DbTerm	**k;
    Node_t *scratcht;
    int scratchi;
    int repeat_count;
    int c;

    if (tb->root == NULL) {
	*ret = am_EOT;
	return DB_ERROR_NONE;
    }
    TOP_NODE(tb, t, i);
    if (t != NULL) {
	term = t->key[i];
	if (! CMP_EQ(key, GETKEY(tb, term->tpl))) {
	    /* Start from the beginning */
	    tb->stack_pos = 0;
	}
    }
    if (tb->stack_pos == 0 || term == NULL) {
	/* Have to rebuild the stack */
	t = tb->root;
	while (t != NULL) {
	    k = t->key; 
	    n = t->cnt;
	    i = binsearch(tb, key, k, n);
	    PUSH_NODE(tb, t, i);
	    if (i < n && CMP_EQ(key, GETKEY(tb, k[i]->tpl))) {
		break;
	    }
	    t = t->ptr[i];
	}
	TOP_NODE(tb, t, i);
	term = t->key[i];
    }

    if (signed_val(key) == 19) {
	erl_printf(CERR, "stack_pos = %d = ", tb->stack_pos);
    }
#ifndef	DEBUG_COMMENT
    {int xxx, idxxxx;
    for (xxx = 0; xxx < tb->stack_pos; xxx++) {
	idxxxx = tb->stack[xxx].idx;
	erl_printf(CERR, "Stack pos %d = ", xxx);
	display(make_tuple(tb->stack[xxx].node->key[idxxxx]->tpl), CERR);
	erl_printf(CERR, "\r\n", xxx);
    }
    }
#endif	/* DEBUG_COMMENT */

    repeat_count = 0;
 repeat:
    if (t == NULL) {
	/* We just popped past the root */
	*ret = am_EOT;
	return DB_ERROR_NONE;
    } else if (i >= t->cnt) {
	/* Move up */
	POP_NODE(tb, t, i);
	if (repeat_count == 0) {
#ifdef	XXX_FLAW_MAYBE
	    /* The first pop just got rid of ourself, need another */
	    POP_NODE(tb, t, i);
#endif	/* XXX_FLAW_MAYBE */
	    /* Got rid of ourself, now read top */
	    TOP_NODE(tb, t, i);
	}
	repeat_count++;
	goto repeat;
    } else if ((c = cmp(key, GETKEY(tb, t->key[i]->tpl))) < 0 &&
	       t->ptr[i] == NULL) {
	/* Want smaller than this leaf, so take this node */
	PUSH_NODE(tb, t, i);
	term = t->key[i];
	/* Fall through */
    } else if (c < 0 && t->ptr[i] != NULL) {
	/* Want smaller than this interior node, so take this node */
	PUSH_NODE(tb, t, i);
	term = t->key[i];
	/* Fall through */
    } else if (t->ptr[i + 1] != NULL && c <= 0) {
	/* Find leftmost child/greatgrandchild/etc. on immediate right */
#ifdef	XXX_WORKS_TINY_NOT_BIGGER
	for (t = t->ptr[i + 1]; t != NULL; t = t->ptr[0]) {
	    PUSH_NODE(tb, t, 0);
	}
	i = 0;
#endif	/* XXX_WORKS_TINY_NOT_BIGGER */
	i = i + 1;
	for (t = t->ptr[i]; t != NULL; t = t->ptr[0]) {
	    PUSH_NODE(tb, t, 0);
	    i = 0;
	}
	TOP_NODE(tb, t, i);
	term = t->key[i];
	/* Fall through */
    } else if (i < t->cnt - 1) {
	/* Choose right sibling in this node */
	POP_NODE(tb, scratcht, scratchi);
	i++;
	PUSH_NODE(tb, t, i);
	term = t->key[i];
	/* Fall through */
    } else {
    repop:
	POP_NODE(tb, t, i);
	if (repeat_count == 0) {
#ifdef	XXX_FLAW_MAYBE
	    /* The first pop just got rid of ourself, need another */
	    POP_NODE(tb, t, i);
#endif	/* XXX_FLAW_MAYBE */
	    /* Got rid of ourself, now read top */
	    TOP_NODE(tb, t, i);
	}
	if (t == NULL) {
	    /* We just popped past the root */
	    *ret = am_EOT;
	    return DB_ERROR_NONE;
	}
	if (t->key[i] == NULL) {
	    repeat_count++;
	    goto repop;
	}
	/* XXX Bogus maybe?? if (CMP_LT(GETKEY(tb, term->tpl), GETKEY(tb, t->key[i]->tpl))) { */
	if (CMP_LT(key, GETKEY(tb, t->key[i]->tpl))) {
	    TOP_NODE(tb, scratcht, scratchi);
	    if (! (scratcht == t && scratchi == i)) { /* Not on stack yet? */
		PUSH_NODE(tb, t, i);
	    }
	    term = t->key[i];
	    /* Fall through */
	} else {
	    repeat_count++;
	    goto repeat;
	}
    }
    /* QQQ deal with bag behavior here! */
    e = GETKEY(tb, term->tpl);
    sz = size_object(e);
    hp = HAlloc(p, sz);
    *ret = copy_struct(e,sz,&hp,&MSO(p));
    return DB_ERROR_NONE;
}
#endif	/* XXX_HOSED */

int
db_prev_btree(Process *p, DbTableBtree *tb, Eterm key, Eterm *ret)
{
    *ret = am_EOT;
    return DB_ERROR_NONE;
}

/* Btree static functions below this point */

/* INSERT_* and DELETE_* are magic constants, do not change! */
#define	INSERT_INCOMPLETE      	0
#define	INSERT_OK		1
#define	INSERT_DUP		2

#define	DELETE_UNDERFLOW	0
#define	DELETE_OK		1
#define	DELETE_NOTFOUND		2

#define	NEWNODE	(safe_alloc_from(41, sizeof(Node_t)))
#define GET_TERM(old, obj)	db_get_term((old), 0, (obj))

static DbTerm *
search(DbTableBtree *tb, Eterm x, Node_t *t)
{
    int		i, j, n;
    DbTerm	**k;

    DBG(("Trace:\n"));
    while (t != NULL) {
	k = t->key; 
	n = t->cnt;
	for (j = 0; j < n; j++) {
	    DBG((" %lu", k[j]));
	}
	DBG(("\n"));
	i = binsearch(tb, x, k, n);
	if (i < n && CMP_EQ(x, GETKEY(tb, k[i]->tpl))) {
	    found(t, i);
	    return k[i];
	}
	t = t->ptr[i];
    }
    notfound(x);
    return NULL;
}

static int
binsearch(DbTableBtree *tb, Eterm x, DbTerm **a, int n)
{
    int i, left, right;

#ifdef	COMMENT
    /*
    ** I tried using the brute-force approach below for n < 8,
    ** but it didn't make any real measurable difference on
    ** large sequential inserts, so I'm commenting it out.
    */
    int xxx = -42;
    /* So few items, don't bother with binary search */
    if (CMP_LE(x, GETKEY(tb, a[0]->tpl))) {
	xxx = 0;
    } else if (CMP_GT(x, GETKEY(tb, a[n-1]->tpl))) {
	xxx = n;
    } else {
	for (i = n - 2; i >= 1; i--) {
	    if (CMP_LE(x, GETKEY(tb, a[i]->tpl))) {
		xxx = i; break;
	    }
	}
    }
    return xxx;
#endif	/* COMMENT */
    if (CMP_LE(x, GETKEY(tb, a[0]->tpl))) {
	return 0;
    }
    if (CMP_GT(x, GETKEY(tb, a[n-1]->tpl))) {
	return n;
    }
    left = 0;
    right = n - 1;
    while (right - left > 1) {
	i = (right + left) >> 1;
	if (CMP_LE(x, GETKEY(tb, a[i]->tpl))) {
	    right = i;
	} else {
	    left = i;
	}
    }
    return right;
}

static void
found(Node_t *t, int i)
{
    int	j;

    DBG(("Found in position %d ", i));
    DBG(("of node with contents: "));
    for (j = 0; j < t->cnt; j++) {
	DBG((" %lu", t->key[j]));
    }
    DBG(("\n"));
}

static void
notfound(Eterm x)
{
    DBG(("Item %lu not found\n", x));
}

static Node_t *
insert(DbTableBtree *tb, Eterm x, Node_t *t)
{
    Node_t	*tnew, *u;
    int		code;
    DbTerm	*xnew;

    code = ins(tb, x, t, &xnew, &tnew);
    if (code == INSERT_DUP) {
	printf("XXX Duplicate key %lu ignored\n", x);
    }
    if (code != INSERT_INCOMPLETE) {
	return t;
    }
    u = NEWNODE;
    u->cnt = 1;
    u->key[0] = xnew;
    u->ptr[0] = t;
    u->ptr[1] = tnew;
    tb->memory += sizeof(Node_t);
    return u;
}

/*
** ins(tb, x, t, y, u) -- Insert x in B-tree with root t.  If not
** completely successful, the key *y and the pointer *u remain to be
** inserted.  Returns INSERT_INCOMPLETE, INSERT_OK, INSERT_DUP.
*/

static int
ins(DbTableBtree *tb, Eterm x, Node_t *t, DbTerm **y, Node_t **u)
{
    Node_t	*tnew, *p_final, **p = t->ptr;
    int		i, j;
    DbTerm	*xnew, *k_final;
    DbTerm	**k = t->key;
    int		*n = &(t->cnt), code;

    /* Is t a pointer field in a leaf? */
    if (t == NULL) {
	/* x is not in the table */
	*u = NULL;
	*y = GET_TERM(NULL, x);	/* Create table's copy of the term */
	tb->memory += (*y)->size;
	tb->nitems++;
	return INSERT_INCOMPLETE;
    }
    /*
    ** Select pointer p[i] and try to insert x in the subtree
    ** of which p[i] is the root.
    */
    i = binsearch(tb, db_getkey(tb->keypos, x), k, *n);
    if (i < *n && CMP_EQ(x, GETKEY(tb, k[i]->tpl))) {
	/* x is indeed present in the table, handle duplicate case here */
	/* XXX Need to implement bag and duplicate_bag behavior here someday */
	/* XXX For now, assume we're a set */
	tb->memory -= (k[i])->size;
	k[i] = GET_TERM(k[i], x);
	tb->memory += (k[i])->size;
	return INSERT_OK;
    }
    code = ins(tb, x, p[i], &xnew, &tnew);
    if (code != INSERT_INCOMPLETE) {
	return code;
    }
    /*
    ** Insertion in subtree did not completely succeed.
    ** Try to insert xnew and tnews in the current node.
    */
    if (*n < MM) {
	i = binsearch(tb, GETKEY(tb, xnew->tpl), k, *n);
	for (j = *n; j > i; j--) {
	    k[j] = k[j-1];
	    p[j+1] = p[j];
	}
	k[i] = xnew;
	p[i+1] = tnew;
	++*n;
	return INSERT_OK;
    }
    /*
    ** The current node was already full, so split it.  Pass item
    ** k[M], in the middle of the augmented sequence, back through
    ** parameter y, so that it can move upward in the tree.  Also,
    ** pass a pointer to the newly-created node back through u.
    ** Return INSERT_INCOMPLETE to report that insertion was not completed.
    */
    if (i == MM) {
	k_final = xnew;
	p_final = tnew;
    } else {
	k_final = k[MM-1];
	p_final = p[MM];
	for (j = MM - 1; j > i; j--) {
	    k[j] = k[j-1];
	    p[j+1] = p[j];
	}
	k[i] = xnew;
	p[i+1] = tnew;
    }
    *y = k[M];
    *n = M;
    *u = NEWNODE;
    (*u)->cnt = M;
    for (j = 0; j < M - 1; j++) {
	(*u)->key[j] = k[j + M + 1];
	(*u)->ptr[j] = p[j + M + 1];
    }
    (*u)->ptr[M-1] = p[MM];
    (*u)->key[M-1] = k_final;
    (*u)->ptr[M] = p_final;
    tb->memory += sizeof(Node_t);
    return INSERT_INCOMPLETE;
}

static Node_t
*delete(DbTableBtree *tb, Eterm x, Node_t *t)
{
    int		code;
    Node_t	*newroot;

    code = del(tb, x, t, t);
    if (code == DELETE_NOTFOUND) {
	DBG(("%lu not found\n", x));
    }
    if (code != DELETE_UNDERFLOW) {
	
	return t;
    }
    /* Underflow: decrease the height of the tree */
    newroot = t->ptr[0];
    sys_free(t);
    tb->memory -= sizeof(Node_t);
    return newroot;
}

static int
del(DbTableBtree *tb, Eterm x, Node_t *t, Node_t *root)
{
    int		i, j, code;
    DbTerm	**k, **item, **lkey, **rkey, **addr, *tmp;
    int		*n, *nleft, *nright, borrowleft, nq;
    Node_t	**p, *left, *right, **lptr, **rptr, *q, *q1;

    if (t == NULL) {
	return DELETE_NOTFOUND;
    }
    n = &(t->cnt);
    k = t->key;
    p = t->ptr;
    i = binsearch(tb, db_getkey(tb->keypos, x), k, *n);
    if (p[0] == NULL) {		/* *t is a leaf */
	if (i == *n || CMP_LT(x, GETKEY(tb, k[i]->tpl))) {
	    return DELETE_NOTFOUND;
	}
	/* x is now equal to k[i], located in a leaf */
	db_free_term_data(k[i]);
	tb->memory -= (k[i])->size;
	tb->nitems--;
	for (j = i + 1; j < *n; j++) {
	    k[j-1] = k[j];
	    p[j] = p[j+1];
	}
	--*n;
	return *n >= ((t == root) ? 1 : M);
    }
    /* *t is an interior node (not a leaf) */
    item = k + i;
    left = p[i];
    nleft = &(left->cnt);
    if (i < *n && CMP_EQ(x, GETKEY(tb, (*item)->tpl))) {
	/*
	** x found in interior node.  Go to left child *p[i] and then
	** follow a path all the way to a leaf, using rightmost
	** branches.
	*/
	q = p[i];
	nq = q->cnt;
	while (q1 = q->ptr[nq], q1 != NULL) {
	    q = q1;
	    nq = q->cnt;
	}
	/*
	** Don't try to free anything now: k[i] is moving to a leaf,
	** it will get removed by recursive del() call.
	*/
	/* Exchange k[i] with the rightmost item in leaf */
	addr = q->key + nq - 1;
	tmp = *item;
	*item = *addr;
	*addr = tmp;
    }
    /* Delete x in subtree with root p[i] */
    code = del(tb, x, left, root);
    if (code != DELETE_UNDERFLOW) {
	return code;
    }
    /* Underflow: borrow, and merge if necessary */
    borrowleft = (i == *n);
    if (borrowleft) {		/* p[i] is rightmost pointer in *p */
	item = k + i - 1;
	left = p[i-1];
	right = p[i];
	nleft = &(left->cnt);
    } else {
	right = p[i+1];
    }
    nright = &(right->cnt);
    lkey = left->key;
    rkey = right->key;
    lptr = left->ptr;
    rptr = right->ptr;
    if (borrowleft) {		/* This is an exception */
	rptr[*nright + 1] = rptr[*nright];
	for (j = *nright; j > 0; j--) {
	    rkey[j] = rkey[j-1];
	    rptr[j] = rptr[j-1];
	}
	++*nright;
	rkey[0] = *item;
	rptr[0] = lptr[*nleft];
	*item = lkey[*nleft - 1];
	if (--*nleft >= M) {
	    return DELETE_OK;
	}
    } else {
	if (*nright > M) {	/* Borrow from right sibling */
	    lkey[M-1] = *item;
	    lptr[M] = rptr[0];
	    *item = rkey[0];
	    ++*nleft;
	    --*nright;
	    for (j = 0; j < *nright; j++) {
		rptr[j] = rptr[j+1];
		rkey[j] = rkey[j+1];
	    }
	    rptr[*nright] = rptr[*nright + 1];
	    return DELETE_OK;
	}
	/* Merge */
	lkey[M-1] = *item;
	lptr[M] = rptr[0];
	for (j = 0; j < M; j++) {
	    lkey[M + j] = rkey[j];
	    lptr[M + j + 1] = rptr[j+1];
	}
	*nleft = MM;
	sys_free(right);
	tb->memory -= sizeof(Node_t);
	for (j = i + 1; j < *n; j++) {
	    k[j-1] = k[j];
	    p[j] = p[j+1];
	}
	return --*n >= ((t == root) ? 1 : M);
    }
    /* NOTREACHED */
    return -666;		/* Shut up the compiler */
}

static unsigned long
print_tree(Node_t *t, int indent_amt, int depth, CIO fd)
{
    int i, j;
    unsigned long count = 0;

    /*
    ** One of the few times we actually want to do a right-side
    ** depth-first traversal.
    */
    if (t->ptr[t->cnt] != NULL) {
	count += print_tree(t->ptr[t->cnt], indent_amt, depth + 1, fd);
    }
    for (i = t->cnt - 1; i >= 0; i--) {
	for (j = 0; j < indent_amt * depth; j++) {
	    erl_printf(fd, " ");
	}
	display(make_tuple(t->key[i]->tpl), fd);
	erl_printf(fd, "\r\n");
	count++;
	if (t->ptr[i] != NULL) {
	    count += print_tree(t->ptr[i], indent_amt, depth + 1, fd);
	}
    }
    return count;
}
